
from .common import (
    BaseChatDownloader,
    Chat,
    Remapper as r,
    Image
)

from ..errors import (
    ChatDownloaderError,
    NoChatReplay,
    ChatDisabled,
    NoContinuation,
    ParsingError,
    VideoUnavailable,
    LoginRequired,
    VideoUnplayable,
    InvalidParameter,
    UserNotFound,
    VideoNotFound,
    NoVideos
)
from ..utils.timed_utils import interruptible_sleep

from ..utils.core import (
    multi_get,
    time_to_seconds,
    seconds_to_time,
    int_or_none,
    float_or_none,
    arbg_int_to_rgba,
    rgba_to_hex,
    try_get_first_key,
    try_get_first_value,
    remove_prefixes,
    remove_suffixes,
    camel_case_split,
    ensure_seconds,
    attempts,
    try_parse_json,
    regex_search,
    parse_iso8601,
    get_title_of_webpage
)

from ..debugging import (log, debug_log)

import json
import time
import random
import re
import hashlib
from requests.exceptions import RequestException
from json.decoder import JSONDecodeError
from urllib import parse


class YouTubeChatDownloader(BaseChatDownloader):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._initialize_consent()

    _NAME = 'youtube.com'

    _SITE_DEFAULT_PARAMS = {
        'format': 'youtube',
    }

    _TESTS = [
        # Get top live streams
        # https://www.youtube.com/results?search_query&sp=CAMSAkAB


        # OTHER:
        # Japanese characters and lots of superchats
        # https://www.youtube.com/watch?v=UlemRwXYWHg

        # strange end times:
        # https://www.youtube.com/watch?v=DzEbfQI4TPQ
        # https://www.youtube.com/watch?v=7PPnCOhkxqo

        # purchased a product linked to the YouTube channel merchandising
        # https://youtu.be/y5ih7nqEoc4


        # TESTING FOR CORRECT FUNCIONALITY
        {
            'name': 'Get chat messages from livestream',
            'params': {
                'url': 'https://www.youtube.com/watch?v=5qap5aO4i9A',
                'timeout': 5
            }
        },
        {
            'name': 'Get chat messages from livestream, using channel id.',
            'params': {
                'url': 'https://www.youtube.com/channel/UCSJ4gkVC6NrvII8umztf0Ow',
                'timeout': 5
            }
        },
        {
            'name': 'Get chat messages from livestream, using custom url (1).',
            'params': {
                'url': 'https://www.youtube.com/c/lofigirl',
                'timeout': 5
            }
        },
        {
            'name': 'Get chat messages from livestream, using custom url (2).',
            'params': {
                'url': 'https://www.youtube.com/lofigirl',
                'timeout': 5
            }
        },
        {
            'name': 'Get chat messages from livestream, using user id.',
            'params': {
                'url': 'https://www.youtube.com/user/YellowBrickCinema',
                'timeout': 5
            }
        },
        {
            'name': 'Get chat messages from live chat replay',
            'params': {
                'url': 'https://www.youtube.com/watch?v=wXspodtIxYU',
                'max_messages': 10
            },

            'expected_result': {
                'message_types': ['text_message'],
                'action_types': ['add_chat_item'],
                'messages_condition': lambda messages: 0 < len(messages) <= 10,
            }
        },
        {
            'name': 'Get top chat messages from live chat replay',
            'params': {
                'url': 'https://www.youtube.com/watch?v=wXspodtIxYU',
                'start_time': 0,
                'end_time': 20,
                'chat_type': 'top'
            },
            'expected_result': {
                'message_types': ['text_message'],
                'action_types': ['add_chat_item'],
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Get superchat and ticker messages from live chat replay',
            'params': {
                'url': 'https://www.youtube.com/watch?v=UlemRwXYWHg',
                'end_time': 20,
                'message_groups': ['superchat', 'tickers']
            },
            'expected_result': {
                'message_types': ['paid_message', 'ticker_paid_message_item', 'membership_item', 'ticker_sponsor_item', 'paid_sticker', 'ticker_paid_sticker_item'],
                'action_types': ['add_chat_item', 'add_live_chat_ticker_item'],
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Get all messages from live chat replay',
            'params': {
                'url': 'https://www.youtube.com/watch?v=97w16cYskVI',
                'end_time': 50,
                'message_types': ['all']
            },
            'expected_result': {
                'message_types': ['viewer_engagement_message', 'paid_message', 'ticker_paid_message_item', 'text_message', 'paid_sticker', 'ticker_paid_sticker_item'],
                'action_types': ['add_chat_item', 'add_live_chat_ticker_item'],
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Get messages from a premiere',  # Premiere
            'params': {
                'url': 'https://www.youtube.com/watch?v=zVCs9Cug_qM',
                'start_time': 0,
                'end_time': 20,
            },
            'expected_result': {
                'message_types': ['text_message'],
                'action_types': ['add_chat_item'],
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Chat replay with donations',
            'params': {
                'url': 'https://www.youtube.com/watch?v=Ih2WTyY62J4',
                'start_time': 0,
                'end_time': 40,
                'message_groups': ['donations']
            },
            'expected_result': {
                'message_types': ['donation_announcement'],
                'action_types': ['add_chat_item'],
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            # 874:24:05 current test
            'name': 'Get chat messages from an unplayable stream.',
            'params': {
                'url': 'https://www.youtube.com/watch?v=V2Afni3S-ok',
                'start_time': 10,
                'end_time': 100,
            },
            'expected_result': {
                'message_types': ['text_message'],
                'action_types': ['add_chat_item'],
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },

        # TESTING FOR ERRORS
        {
            'name': 'Video does not exist',
            'params': {
                'url': 'https://www.youtube.com/watch?v=xxxxxxxxxxx',
            },
            'expected_result': {
                'error': VideoUnavailable,
            }
        },
        {
            'name': 'This video is no longer available due to a copyright claim by International Olympic Committee.',
            'params': {
                'url': 'https://www.youtube.com/watch?v=cjk2UKkzY0g',
            },
            'expected_result': {
                'error': VideoUnavailable,
            }
        },
        {
            'name': 'This video is not available.',  # YouTube Premium
            'params': {
                'url': 'https://www.youtube.com/watch?v=i1Ko8UG-Tdo',
            },
            'expected_result': {
                'error': VideoUnplayable,
            }
        },
        {
            'name': 'This video is not available.',  # Rental video preview
            'params': {
                'url': 'https://www.youtube.com/watch?v=yYr8q0y5Jfg',
            },
            'expected_result': {
                'error': VideoUnplayable,
            }
        },
        {
            # The following content has been identified by the YouTube community
            # as inappropriate or offensive to some audiences.
            'name': "This video has been removed for violating YouTube's policy on hate speech. Learn more about combating hate speech in your country.",  # Rental video preview
            'params': {
                'url': 'https://www.youtube.com/watch?v=6SJNVb0GnPI',
            },
            'expected_result': {
                'error': VideoUnavailable,
            }
        },
        {
            'name': 'Members-only content',
            'params': {
                'url': 'https://www.youtube.com/watch?v=vprErlL1w2E',
            },
            'expected_result': {
                'error': VideoUnplayable,
            }
        },
        {
            'name': 'Chat is disabled for this live stream',
            'params': {
                'url': 'https://www.youtube.com/watch?v=XWq5kBlakcQ',
            },
            'expected_result': {
                'error': ChatDisabled,
            }
        },
        {
            'name': 'Live chat replay has been turned off for this video',
            'params': {
                'url': 'https://www.youtube.com/watch?v=7lGZvbasx6A',
            },
            'expected_result': {
                'error': NoChatReplay,
            }
        },

        {
            'name': 'Video is private',
            'params': {
                'url': 'https://www.youtube.com/watch?v=ijFMXqa-N0c',
            },
            'expected_result': {
                'error': LoginRequired,
            }
        },
        {
            'name': 'The uploader has not made this video available in your country.',
            'params': {
                'url': 'https://www.youtube.com/watch?v=sJL6WA-aGkQ',
            },
            'expected_result': {
                'error': VideoUnplayable,
            }
        },
        {
            'name': 'This live stream recording is not available.',
            'params': {
                'url': 'https://www.youtube.com/watch?v=qEJwOuvDf7I',
            },
            'expected_result': {
                'error': VideoUnplayable,
            }
        },
        {
            # Age restricted
            'name': 'Sign in to confirm your age. This video may be inappropriate for some users.',
            'params': {
                'url': 'https://www.youtube.com/watch?v=WaOKSUlf4TM',
            },
            'expected_result': {
                'error': LoginRequired,
            }
        },

        # Potential parsing errors
        {
            'name': "Parsing error with '};' inside yt initial data (1)",
            'params': {
                'url': 'https://www.youtube.com/watch?v=CHqg6qOn4no',
            },
            'expected_result': {
                'error': NoChatReplay,
            }
        },
        {
            'name': "Parsing error with '};' inside yt initial data (2)",
            'params': {
                'url': 'https://www.youtube.com/watch?v=gVfgbahppCY',
            },
            'expected_result': {
                'error': NoChatReplay,
            }
        },
        {
            'name': 'Title with JS-like syntax "};"',
            'params': {
                'url': 'https://www.youtube.com/watch?v=lsguqyKfVQg',
            },
            'expected_result': {
                'error': NoChatReplay,
            }
        },

        # Clips
        {
            'name': 'Chat replay of clip (past broadcast)',
            'params': {
                'url': 'https://www.youtube.com/clip/Ugy_1IfsnZUWZSXL6C94AaABCQ',
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Chat replay of clip (premiere)',
            'params': {
                'url': 'https://www.youtube.com/clip/UgzNZCNnPzq-M3_Utjl4AaABCQ',
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Clip does not have a chat replay.',
            'params': {
                'url': 'https://www.youtube.com/clip/UgwVu73xQ5FUiGnteZJ4AaABCQ',
            },
            'expected_result': {
                'error': NoChatReplay,
            }
        },
        {
            'name': 'Sign in to confirm your age. This clip may be inappropriate for some users.',
            'params': {
                'url': 'https://www.youtube.com/clip/UgyfnqwleyOmnO-qA1h4AaABCQ',
            },
            'expected_result': {
                'error': LoginRequired,
            }
        },
        {
            'name': "Clip not available. The clip can be unavailable if it was deleted, or if the video it's based on was removed or edited.",
            'params': {
                'url': 'https://youtube.com/clip/UgxJiPo-4EeSYDfrYp94AaABCQ',
            },
            'expected_result': {
                'error': VideoUnavailable,
            }
        },
        {
            'name': 'Clip does not exist',
            'params': {
                'url': 'https://youtube.com/clip/x',
            },
            'expected_result': {
                'error': VideoNotFound,
            }
        }
    ]

    _YT_INITIAL_BOUNDARY_RE = r'\s*(?:var\s+(?:meta|head)|</script|\n)'
    _YT_INITIAL_DATA_RE = r'(?:window\s*\[\s*["\']ytInitialData["\']\s*\]|ytInitialData)\s*=\s*({.+?})\s*;' + \
        _YT_INITIAL_BOUNDARY_RE
    _YT_INITIAL_PLAYER_RESPONSE_RE = r'ytInitialPlayerResponse\s*=\s*({.+?})\s*;' + \
        _YT_INITIAL_BOUNDARY_RE
    _YT_CFG_RE = r'ytcfg\.set\s*\(\s*({.+?})\s*\)\s*;'

    _YT_HOME = 'https://www.youtube.com'
    _YT_VIDEO_TEMPLATE = _YT_HOME + '/watch?v={}'
    _YT_CLIP_TEMPLATE = _YT_HOME + '/clip/{}'

    _YOUTUBE_INIT_API_TEMPLATE = _YT_HOME + '/{}?continuation={}'
    _YOUTUBE_CHAT_API_TEMPLATE = _YT_HOME + '/youtubei/v1/live_chat/get_{}?key={}'
    _YOUTUBE_BROWSE_API_TEMPLATE = _YT_HOME + '/youtubei/v1/browse?key={}'

    _MESSAGE_GROUPS = {
        'messages': [
            'text_message'  # normal message
        ],
        'superchat': [
            # superchat messages which appear in chat
            'membership_item',
            'paid_message',
            'paid_sticker',
        ],
        'tickers': [
            # superchat messages which appear ticker (at the top)
            'ticker_paid_sticker_item',
            'ticker_paid_message_item',
            'ticker_sponsor_item',
        ],
        'banners': [
            'banner',
            'banner_header'
        ],

        'donations': [
            'donation_announcement'
        ],
        'engagement': [
            # message saying live chat replay is on
            'viewer_engagement_message',
        ],
        'purchases': [
            'purchased_product_message'  # product purchased
        ],

        'mode_changes': [
            'mode_change_message'  # e.g. slow mode enabled
        ],

        'deleted': [
            'deleted_message'
        ],
        'bans': [
            'ban_user'
        ],

        'placeholder': [
            'placeholder_item'  # placeholder
        ]
    }

    _MESSAGE_TYPES = ['all']
    for _group in _MESSAGE_GROUPS:
        _MESSAGE_TYPES += _MESSAGE_GROUPS[_group]

    # Regex provided by youtube-dl

    # TODO differentiate /user/x, /c/y and /channel/UC...

    _VALID_URLS = {
        '_get_chat_by_video_id': r'''(?x)^
                     (
                         # http(s):// or protocol-independent URL
                         (?:https?://|//)
                         (?:(?:(?:(?:\w+\.)?[yY][oO][uU][tT][uU][bB][eE](?:-nocookie|kids)?\.com/|
                            youtube\.googleapis\.com/)                        # the various hostnames, with wildcard subdomains
                         (?:.*?\#/)?                                          # handle anchor (#/) redirect urls
                         (?:                                                  # the various things that can precede the ID:
                             # v/ or embed/ or e/
                             (?:(?:v|embed|e)/(?!videoseries))
                             |(?:                                             # or the v= param in all its forms
                                 # preceding watch(_popup|.php) or nothing (like /?v=xxxx)
                                 (?:(?:watch|movie)(?:_popup)?(?:\.php)?/?)?
                                 (?:\?|\#!?)                                  # the params delimiter ? or # or #!
                                 # any other preceding param (like /?s=tuff&v=xxxx or ?s=tuff&amp;v=V36LpHqtcDY)
                                 (?:.*?[&;])??
                                 v=
                             )
                         ))
                         |(?:
                            youtu\.be                                        # just youtu.be/xxxx
                         )/)
                     )?                                                       # all until now is optional -> you can pass the naked ID
                     # here is it! the YouTube video ID
                     (?P<id>[0-9A-Za-z_-]{11})''',

        '_get_chat_by_clip_id': r'''(?x)
                (?:https?://|//)
                    (?:\w+\.)?
                    (?:
                        youtube?\.com
                    )/clip/
                    (?P<id>[a-zA-Z0-9_-]+)''',

        # while this does match 'watch' urls, it will never
        # return this since the above regex is run before this
        '_get_chat_by_user': r'''(?x)
                (?:https?://|//)
                    (?:\w+\.)?
                    (?:
                        youtube(?:kids)?\.com
                    )/
                    (?:
                        (?P<type>channel|c|user)/
                    )?
                    (?P<id>[a-zA-Z0-9_-]+)'''
    }

    @ staticmethod
    def _get_source_image_url(url):
        index = url.find('=')
        if index >= 0:
            return url[0:url.index('=')]
        else:
            return url

    @ staticmethod
    def _parse_youtube_link(text):
        if text.startswith(('/redirect', 'https://www.youtube.com/redirect')):  # is a redirect link
            info = dict(parse.parse_qsl(parse.urlsplit(text).query))
            return info.get('q') or ''
        elif text.startswith('//'):
            return 'https:' + text
        elif text.startswith('/'):  # is a youtube link e.g. '/watch','/results'
            return YouTubeChatDownloader._YT_HOME + text
        else:  # is a normal link
            return text

    @ staticmethod
    def _parse_navigation_endpoint(navigation_endpoint, default_text=''):
        try:
            return YouTubeChatDownloader._parse_youtube_link(
                navigation_endpoint['commandMetadata']['webCommandMetadata']['url'])
        except Exception:
            return default_text

    @ staticmethod
    def _parse_text(info):
        return YouTubeChatDownloader._parse_runs(info)['message'] or YouTubeChatDownloader._get_simple_text(info)

    @ staticmethod
    def _parse_runs(run_info, parse_links=True):
        """ Reads and parses YouTube formatted messages (i.e. runs). """

        # TODO separate _parse_runs logic?

        message_info = {
            'message': ''
        }

        if not isinstance(run_info, dict):
            return message_info

        message_emotes = {}

        runs = run_info.get('runs') or []
        for run in runs:
            if 'text' in run:
                if parse_links and 'navigationEndpoint' in run:  # is a link and must parse

                    # if something fails, use default text
                    message_info['message'] += YouTubeChatDownloader._parse_navigation_endpoint(
                        run['navigationEndpoint'], run['text'])

                else:  # is a normal message
                    message_info['message'] += run['text']

            elif 'emoji' in run:
                emoji = run['emoji']
                emoji_id = emoji.get('emojiId')

                name = multi_get(emoji, 'shortcuts', 0)

                if name:
                    if emoji_id and emoji_id not in message_emotes:

                        # TODO change to remapping?
                        message_emotes[emoji_id] = {
                            'id': emoji_id,
                            'name': name,
                            'shortcuts': emoji.get('shortcuts'),
                            'search_terms': emoji.get('searchTerms'),
                            'images': YouTubeChatDownloader._parse_thumbnails(emoji.get('image', {})),
                            'is_custom_emoji': emoji.get('isCustomEmoji', False)
                        }

                    message_info['message'] += name

            else:
                # unknown run
                message_info['message'] += str(run)

        if message_emotes:
            message_info['emotes'] = list(message_emotes.values())

        return message_info

    @ staticmethod
    def _parse_item(item, info=None, offset=0):
        if info is None:
            info = {}
        # info is starting point
        item_index = try_get_first_key(item)
        item_info = item.get(item_index)

        if not item_info:
            return info

        for key in item_info:
            r.remap(info, YouTubeChatDownloader._REMAPPING,
                    key, item_info[key])

        # check for colour information
        for colour_key in YouTubeChatDownloader._COLOUR_KEYS:
            if colour_key in item_info:  # if item has colour information
                rgba_colour = arbg_int_to_rgba(item_info[colour_key])
                hex_colour = rgba_to_hex(rgba_colour)
                new_key = camel_case_split(
                    colour_key.replace('Color', 'Colour'))
                info[new_key] = hex_colour

        item_endpoint = item_info.get('showItemEndpoint')
        if item_endpoint:  # has additional information
            renderer = multi_get(
                item_endpoint, 'showLiveChatItemEndpoint', 'renderer')

            if renderer:
                info.update(YouTubeChatDownloader._parse_item(
                    renderer, offset=offset))

        BaseChatDownloader._move_to_dict(info, 'author')

        # TODO determine if youtube glitch has occurred
        # round(time_in_seconds/timestamp) == 1
        time_in_seconds = info.get('time_in_seconds')
        time_text = info.get('time_text')

        if time_in_seconds is not None:

            if time_text is not None:
                # All information was provided, check if time_in_seconds is <= 0
                # For some reason, YouTube sets the video offset to 0 if the message
                # was sent before the stream started. This fixes that:
                if time_in_seconds <= 0:
                    info['time_in_seconds'] = time_to_seconds(time_text)
            else:
                # recreate time text from time in seconds
                info['time_text'] = seconds_to_time(time_in_seconds)

        elif time_text is not None:  # doesn't have time in seconds, but has time text
            info['time_in_seconds'] = time_to_seconds(time_text)
        else:
            pass
            # has no current video time information
            # (usually live video or a sub-item)

        # non-zero, non-null offset and has time_in_seconds info
        if offset and 'time_in_seconds' in info:
            info['time_in_seconds'] -= offset
            info['time_text'] = seconds_to_time(info['time_in_seconds'])

        if 'message' not in info:  # Ensure the parsed item contains the 'message' key
            info['message'] = None

        return info

    @ staticmethod
    def _parse_badges(badge_items):
        badges = []

        for badge in badge_items:
            to_add = {}
            parsed_badge = YouTubeChatDownloader._parse_item(badge)

            title = parsed_badge.pop('tooltip', None)
            if title:
                to_add['title'] = title

            icon = parsed_badge.pop('icon', None)
            if icon:
                to_add['icon_name'] = icon.lower()

            badge_icons = parsed_badge.pop('badge_icons', None)
            if badge_icons:
                to_add['icons'] = []

                url = None
                for icon in badge_icons:
                    url = icon.get('url')
                    if url:
                        matches = re.search(r'=s(\d+)', url)
                        if matches:
                            size = int(matches.group(1))
                            to_add['icons'].append(
                                Image(url, size, size).json())
                if url:
                    to_add['icons'].insert(0, Image(
                        YouTubeChatDownloader._get_source_image_url(url), image_id='source').json())

            badges.append(to_add)

            # if 'member'
            # remove the tooltip afterwards
            # print(badges)
        return badges

    @ staticmethod
    def _parse_thumbnails(item):

        # sometimes thumbnails come as a list
        if isinstance(item, list):
            item = item[0]  # rebase

        # TODO add source:
        # https://yt3.ggpht.com/ytc/AAUvwnhBYeK7_iQTJbXe6kIMpMlCI2VsVHhb6GBJuYeZ=s32-c-k-c0xffffffff-no-rj-mo
        # https://yt3.ggpht.com/ytc/AAUvwnhBYeK7_iQTJbXe6kIMpMlCI2VsVHhb6GBJuYeZ

        thumbnails = item.get('thumbnails') or []
        final = list(map(lambda x: Image(**x).json(), thumbnails))

        if len(final) > 0:
            final.insert(0, Image(
                YouTubeChatDownloader._get_source_image_url(final[0]['url']), image_id='source').json())

        return final

    @ staticmethod
    def _parse_action_button(item):
        endpoint = multi_get(item, 'buttonRenderer', 'navigationEndpoint')

        return {
            'url': YouTubeChatDownloader._parse_navigation_endpoint(endpoint) if endpoint else '',
            'text': multi_get(item, 'buttonRenderer', 'text', 'simpleText') or ''
        }

    @ staticmethod
    def _get_simple_text(item):
        return item.get('simpleText')

    _CURRENCY_SYMBOLS = {
        '$': 'USD',
        'A$': 'AUD',
        'CA$': 'CAD',
        'HK$': 'HKD',
        'MX$': 'MXN',
        'NT$': 'TWD',
        'NZ$': 'NZD',
        'R$': 'BRL',
        '£': 'GBP',
        '€': 'EUR',
        '₹': 'INR',
        '₪': 'ILS',
        '₱': 'PHP',

        '₩': 'KRW',
        '￦': 'KRW',

        '¥': 'JPY',
        '￥': 'JPY',
    }

    # All other currency symbols use the ISO 4217 format:
    # https://en.wikipedia.org/wiki/ISO_4217
    # e.g. 'CHF', 'COP', 'HUF', 'PLN', 'RUB', 'SEK', 'PEN', 'ARS', 'CLP', 'NOK', 'BAM', 'SGD'

    @ staticmethod
    def _parse_currency(item):
        mixed_text = item.get('simpleText') or str(item)

        info = re.split(r'([\d,\.]+)', mixed_text)
        if len(info) >= 2:  # Correct parse
            currency_symbol = info[0].strip()
            currency_code = YouTubeChatDownloader._CURRENCY_SYMBOLS.get(
                currency_symbol, currency_symbol)
            amount = float(info[1].replace(',', ''))

        else:  # Unable to get info
            amount = float(re.sub(r'[^\d\.]+', '', mixed_text))
            currency_symbol = currency_code = None

        return {
            'text': mixed_text,
            'amount': amount,
            'currency': currency_code,  # ISO_4217
            'currency_symbol': currency_symbol
        }

    _REMAPPING = {
        'id': 'message_id',
        'authorExternalChannelId': 'author_id',
        'authorName': r('author_name', _get_simple_text),
        # TODO author_display_name
        'purchaseAmountText': r('money', _parse_currency),
        'message': r(None, _parse_runs, True),
        'timestampText': r('time_text', _get_simple_text),
        'timestampUsec': r('timestamp', int_or_none),

        'authorPhoto': r('author_images', _parse_thumbnails),

        'tooltip': 'tooltip',

        'icon': r('icon', lambda x: x.get('iconType')),
        'authorBadges': r('author_badges', _parse_badges),

        # stickers
        'sticker': r('sticker_images', _parse_thumbnails),

        # ticker_paid_message_item
        'fullDurationSec': r('ticker_duration', int_or_none),
        'amount': r('money', _parse_currency),


        # ticker_sponsor_item
        'detailText': r(None, _parse_runs, True),
        'customThumbnail': r('badge_icons', _parse_thumbnails),

        # membership_item
        'headerPrimaryText': r('header_primary_text', _parse_text),
        'headerSubtext': r('header_secondary_text', _parse_text),
        'sponsorPhoto': r('sponsor_icons', _parse_thumbnails),

        # ticker_paid_sticker_item
        'tickerThumbnails': r('ticker_icons', _parse_thumbnails),

        # deleted messages
        'deletedStateMessage': r(None, _parse_runs, True),
        'targetItemId': 'target_message_id',

        'externalChannelId': 'author_id',

        # action buttons
        'actionButton': r('action', _parse_action_button),

        # addBannerToLiveChatCommand
        'text': r(None, _parse_runs, True),
        'viewerIsCreator': 'viewer_is_creator',
        'targetId': 'target_message_id',
        'isStackable': 'is_stackable',
        'backgroundType': 'background_type',

        # removeBannerForLiveChatCommand
        'targetActionId': 'target_message_id',

        # donation_announcement
        'subtext': r(None, _parse_runs, True),

        # tooltip
        'detailsText': r(None, _parse_runs, True),

    }

    _COLOUR_KEYS = [
        # paid_message
        'authorNameTextColor', 'timestampColor', 'bodyBackgroundColor',
        'headerTextColor', 'headerBackgroundColor', 'bodyTextColor',

        # paid_sticker
        'backgroundColor', 'moneyChipTextColor', 'moneyChipBackgroundColor',

        # ticker_paid_message_item
        'startBackgroundColor', 'amountTextColor', 'endBackgroundColor',

        # ticker_sponsor_item
        'detailTextColor'
    ]

    _STICKER_KEYS = [
        # to actually ignore
        'stickerDisplayWidth', 'stickerDisplayHeight',  # ignore

        # parsed elsewhere
        'sticker',
    ]

    _KEYS_TO_IGNORE = [
        # to actually ignore
        'contextMenuAccessibility', 'contextMenuEndpoint', 'trackingParams', 'accessibility', 'dwellTimeMs',

        'empty',  # signals liveChatMembershipItemRenderer has no message body

        'contextMenuButton',

        # parsed elsewhere
        'showItemEndpoint',
        'durationSec',

        # banner parsed elsewhere
        'header', 'contents', 'actionId',

        # tooltipRenderer
        'dismissStrategy', 'suggestedPosition', 'promoConfig'
    ]

    _KNOWN_KEYS = set(list(_REMAPPING.keys()) +
                      _COLOUR_KEYS + _STICKER_KEYS + _KEYS_TO_IGNORE)

    # KNOWN ACTIONS AND MESSAGE TYPES
    _KNOWN_ADD_TICKER_TYPES = {
        'addLiveChatTickerItemAction': [
            'liveChatTickerSponsorItemRenderer',
            'liveChatTickerPaidStickerItemRenderer',
            'liveChatTickerPaidMessageItemRenderer'
        ]
    }

    _KNOWN_ADD_ACTION_TYPES = {
        'addChatItemAction': [
            # message saying Live Chat replay is on
            'liveChatViewerEngagementMessageRenderer',
            'liveChatMembershipItemRenderer',
            'liveChatTextMessageRenderer',
            'liveChatPaidMessageRenderer',
            'liveChatPlaceholderItemRenderer',  # placeholder
            'liveChatDonationAnnouncementRenderer',

            'liveChatPaidStickerRenderer',
            'liveChatModeChangeMessageRenderer',  # e.g. slow mode enabled

            # TODO find examples of:
            # 'liveChatPurchasedProductMessageRenderer',  # product purchased
            # liveChatLegacyPaidMessageRenderer
            # liveChatModerationMessageRenderer
            # liveChatAutoModMessageRenderer

        ]
    }

    _KNOWN_REPLACE_ACTION_TYPES = {
        'replaceChatItemAction': [
            'liveChatPlaceholderItemRenderer',
            'liveChatTextMessageRenderer'
        ]
    }

    # actions that have an 'item'
    _KNOWN_ITEM_ACTION_TYPES = {
        **_KNOWN_ADD_TICKER_TYPES, **_KNOWN_ADD_ACTION_TYPES}

    # [message deleted] or [message retracted]
    _KNOWN_REMOVE_ACTION_TYPES = {
        'markChatItemsByAuthorAsDeletedAction': [  # TODO ban?
            'banUser'  # deletedStateMessage
        ],
        'markChatItemAsDeletedAction': [
            'deletedMessage'  # deletedStateMessage
        ]
    }

    _KNOWN_ADD_BANNER_TYPES = {
        'addBannerToLiveChatCommand': [
            'liveChatBannerRenderer',
            'liveChatBannerHeaderRenderer',
            'liveChatTextMessageRenderer',
        ]
    }

    _KNOWN_REMOVE_BANNER_TYPES = {
        'removeBannerForLiveChatCommand': [
            'removeBanner'  # targetActionId
        ]
    }

    _KNOWN_TOOLTIP_ACTION_TYPES = {
        'showLiveChatTooltipCommand': [
            'tooltipRenderer'
        ]
    }

    # Not come across yet (actions/commands)
    # search "livechat"
    # https://github.com/reachomk/ytvanced/tree/master/src/main/java/com/google/protos/youtube/api/innertube
    # addLiveChatTextMessageFromTemplateAction
    # liveChatMessageBuyFlowHeadingRenderer
    # liveChatPaidMessageFooterRenderer
    # liveChatProductButtonRenderer
    # liveChatPurchaseMessageEndpoint
    # removeChatItemAction
    # replaceLiveChatRendererAction
    # showLiveChatDialogAction
    # showLiveChatSurveyCommand

    # Not checked for
    # _KNOWN_IGNORE_ACTION_TYPES = {
    #     'authorBadges': [
    #         'liveChatAuthorBadgeRenderer'
    #     ],
    #     'showLiveChatItemEndpoint': [
    #         'liveChatPaidStickerRenderer',
    #         'liveChatPaidMessageRenderer',
    #         'liveChatMembershipItemRenderer'
    #     ]
    # }

    _KNOWN_POLL_ACTION_TYPES = {
    }

    _KNOWN_IGNORE_ACTION_TYPES = {

        # TODO add support for poll actions
        'showLiveChatActionPanelAction': [],
        'updateLiveChatPollAction': [],
        'closeLiveChatActionPanelAction': []

    }

    _KNOWN_ACTION_TYPES = {
        **_KNOWN_ITEM_ACTION_TYPES,
        **_KNOWN_REMOVE_ACTION_TYPES,
        **_KNOWN_REPLACE_ACTION_TYPES,
        **_KNOWN_ADD_BANNER_TYPES,
        **_KNOWN_REMOVE_BANNER_TYPES,
        **_KNOWN_TOOLTIP_ACTION_TYPES,
        **_KNOWN_POLL_ACTION_TYPES,
        **_KNOWN_IGNORE_ACTION_TYPES
    }

    _KNOWN_IGNORE_MESSAGE_TYPES = [
        'liveChatPlaceholderItemRenderer'
    ]
    _KNOWN_MESSAGE_TYPES = []
    for _action in _KNOWN_ACTION_TYPES:
        _KNOWN_MESSAGE_TYPES += _KNOWN_ACTION_TYPES[_action]

    _KNOWN_SEEK_CONTINUATIONS = [
        'playerSeekContinuationData'
    ]

    _KNOWN_CHAT_CONTINUATIONS = [
        'invalidationContinuationData', 'timedContinuationData',
        'liveChatReplayContinuationData', 'reloadContinuationData'
    ]

    _KNOWN_CONTINUATIONS = _KNOWN_SEEK_CONTINUATIONS + _KNOWN_CHAT_CONTINUATIONS

    def generate_urls(self, **kwargs):
        items = self._get_testing_items()

        for item in items:
            yield self._YT_VIDEO_TEMPLATE.format(item['video_id'])

        # downloader.get_playlist_items

    _LIVE_PLAYLIST_URL = _YT_HOME + '/channel/UC4R8DWoMoI7CAwX8_LjQHig'

    def _get_testing_items(self):
        params = {
            'max_attempts': 10
        }

        yt_initial_data, ytcfg, player_response_info = self._get_initial_info(
            self._LIVE_PLAYLIST_URL, params)

        sections = yt_initial_data['contents']['twoColumnBrowseResultsRenderer']['tabs'][
            0]['tabRenderer']['content']['sectionListRenderer']['contents']

        for section in sections:
            section_info = section['itemSectionRenderer']['contents'][0]['shelfRenderer']

            # print(section_info)

            # section_title = section_info['title']['runs'][0]['text']
            # print(section_title)

            # items = section_info['content']['horizontalListRenderer']['items']

            playlist_url = self._YT_HOME + \
                section_info['endpoint']['commandMetadata']['webCommandMetadata']['url']

            yield from self.get_playlist_items(playlist_url)

    @ staticmethod
    def _get_rendered_content(yt_info, tab_index=0):
        return yt_info['contents']['twoColumnBrowseResultsRenderer']['tabs'][tab_index]['tabRenderer']['content'][
            'sectionListRenderer']['contents'][0]['itemSectionRenderer']['contents'][0]

    _VIDEO_REMAPPING = {
        'videoId': 'video_id',
        'title': r('title', lambda x: YouTubeChatDownloader._parse_runs(x)['message']),
        'viewCountText': r('view_count', lambda x: YouTubeChatDownloader._parse_text(x)),
        'shortViewCountText': r('short_view_count', lambda x: YouTubeChatDownloader._parse_text(x)),

        # 'videoId', 'thumbnail', 'title', 'viewCountText', 'navigationEndpoint', 'ownerBadges', 'trackingParams', 'shortViewCountText', 'menu', 'thumbnailOverlays'
    }

    @ staticmethod
    def _parse_video(video_renderer):
        return r.remap_dict(video_renderer, YouTubeChatDownloader._VIDEO_REMAPPING)

    _VIDEO_STATUS_REMAPPING = {
        'all': 'all',
        'live': (501, 'Live now'),
        'upcoming': (502, 'Upcoming live streams'),
        'past': (503, 'Past live streams')
    }

    def get_user_videos(self, channel_id=None, user_id=None, custom_username=None, video_status='all', params=None):
        """Retrieve all videos listed on the user's channel

        If more than one of `channel_id`, `user_id` and `custom_username`
        are specifed, the first one specified will be returned.

        :param channel_id: The user's channel ID, defaults to None.
            (e.g., https://www.youtube.com/channel/<channel_id>)
        :type channel_id: str, optional
        :param user_id: The user's ID, defaults to None
            (e.g., https://www.youtube.com/user/<user_id>)
        :type user_id: str, optional
        :param custom_username: [description], defaults to None
            (e.g., https://www.youtube.com/c/<custom_username>)
        :type custom_username: str, optional
        :param video_status: Determines which videos will be retrieved, defaults to 'all'.
            Must be one of 'all', 'live', 'upcoming' or 'past'.
        :type video_status: str, optional
        :param params: Additional program parameters, defaults to None
        :type params: dict, optional
        :raises ValueError: If no user is specified or an invalid video_status is specified
        :raises UserNotFound: If the user cannot be found
        :raises NoVideos: If the channel has no videos
        :yield: The next video
        :rtype: dict
        """

        _id = ''
        _type = ''
        if channel_id:
            _id = channel_id
            _type = 'channel'
        elif user_id:
            _id = user_id
            _type = 'user'
        elif custom_username:
            _id = custom_username
            _type = 'c'
        else:
            raise ValueError('No user type specified.')

        # live, past, upcoming
        vid_type = self._VIDEO_STATUS_REMAPPING.get(video_status.lower())

        if not vid_type:
            raise ValueError(
                f'Invalid argument passed for video_status. Must be one of {set(self._VIDEO_STATUS_REMAPPING.keys())}')

        user_url = f'https://www.youtube.com/{_type}/{_id}'

        if vid_type == 'all':
            vids_url = f'{user_url}/videos'
        else:
            vids_url = f'{user_url}/videos?view=2&live_view={vid_type[0]}'

        yt_info, ytcfg, _ = self._get_initial_info(vids_url, params)

        section_list_renderer = multi_get(
            yt_info, 'contents', 'twoColumnBrowseResultsRenderer', 'tabs', 1, 'tabRenderer', 'content', 'sectionListRenderer')
        if not section_list_renderer:
            raise UserNotFound(f'Unable to find user: "{user_url}"')

        sub_menu_items = multi_get(
            section_list_renderer, 'subMenu', 'channelSubMenuRenderer', 'contentTypeSubMenuItems')
        if not sub_menu_items:
            raise NoVideos('This channel has no videos.')

        # Check that the returned grid is what was asked for
        # YouTube tries to correct your mistake by selecting the uploads tab
        # if you try to access a tab that is not visible.
        selected = list(filter(lambda x: x['selected'], sub_menu_items))
        if vid_type != 'all' and (not selected or selected[0]['title'] != vid_type[1]):
            log('debug',
                f'"{vid_type[1]}" tab is not visible for this channel (i.e. there are no such videos).')
            return

        api_key = ytcfg.get('INNERTUBE_API_KEY')
        continuation_url = self._YOUTUBE_BROWSE_API_TEMPLATE.format(api_key)

        # innertube_context =
        # print('innertube_context', innertube_context)
        continuation_params = {
            'context': ytcfg.get('INNERTUBE_CONTEXT') or {}
        }
        continuation = None
        first_time = True
        while True:
            if first_time:
                items = multi_get(section_list_renderer, 'contents', 0,
                                  'itemSectionRenderer', 'contents', 0, 'gridRenderer', 'items')
                first_time = False
            else:
                continuation_params['continuation'] = continuation
                yt_info = self._get_continuation_info(
                    continuation_url, params, json=continuation_params)
                items = multi_get(yt_info, 'onResponseReceivedActions',
                                  0, 'appendContinuationItemsAction', 'continuationItems')

            if not items:
                break

            continuation = None
            for item in items:
                vid = item.get('gridVideoRenderer')
                continuation_item = item.get('continuationItemRenderer')

                if vid:
                    yield self._parse_video(vid)
                elif continuation_item:
                    continuation = multi_get(
                        continuation_item, 'continuationEndpoint', 'continuationCommand', 'token')

            if not continuation:
                break

    def get_playlist_items(self, playlist_url, params=None):

        yt_initial_data, ytcfg, player_response_info = self._get_initial_info(
            playlist_url, params)

        items = self._get_rendered_content(
            yt_initial_data)['playlistVideoListRenderer']['contents']

        for item in items:
            playlist_video = item.get('playlistVideoRenderer')

            if playlist_video:
                yield self._parse_video(playlist_video)

    _CONSENT_ID_REGEX = r'PENDING\+(\d+)'
    # https://github.com/ytdl-org/youtube-dl/blob/a8035827177d6b59aca03bd717acb6a9bdd75ada/youtube_dl/extractor/youtube.py#L251

    def _initialize_consent(self):
        if self.get_cookie_value('__Secure-3PSID'):
            return

        consent_id = None
        consent = self.get_cookie_value('CONSENT')

        if consent:
            if 'YES' in consent:
                return

            consent_id = regex_search(consent, self._CONSENT_ID_REGEX)

        if not consent_id:
            consent_id = random.randint(100, 999)

        self.set_cookie_value('.youtube.com', 'CONSENT',
                              f'YES+cb.20210328-17-p0.en+FX+{consent_id}')

    def _generate_sapisidhash_header(self):
        sapis_id = self.get_cookie_value('SAPISID')
        sapisid_cookie = self.get_cookie_value('__Secure-3PAPISID') or sapis_id

        if sapisid_cookie is None:
            return

        time_now = round(time.time())

        # SAPISID cookie is required if not already present
        if not sapis_id:
            self.set_cookie_value(
                '.youtube.com', 'SAPISID', sapisid_cookie, secure=True, expire_time=time_now + 3600)

        # SAPISIDHASH algorithm from https://stackoverflow.com/a/32065323
        sapisidhash = hashlib.sha1(
            f'{time_now} {sapisid_cookie} {self._YT_HOME}'.encode('utf-8')).hexdigest()
        return f'SAPISIDHASH {time_now}_{sapisidhash}'

    def _get_continuation_info(self, continuation_url, program_params, **post_kwargs):
        if program_params is None:
            program_params = {}
        max_attempts = program_params.get('max_attempts', 1)

        for attempt_number in attempts(max_attempts):
            try:
                response = self._session_post(continuation_url, **post_kwargs)
                json_response = response.json()

                # Check for errors:
                error = json_response.get('error')
                if error:
                    error_code = error.get('code')
                    error_message = error.get('message')

                    if error_code // 100 == 5:  # Server error, retry
                        self.retry(attempt_number,
                                   text=error_message, **program_params)
                        continue
                    # 404 means deleted while live

                return json_response

            except JSONDecodeError as e:
                self.retry(attempt_number, error=e, **program_params,
                           text=f'Unable to parse JSON: `{response.text}`')

            except RequestException as e:
                self.retry(attempt_number, error=e, **program_params)

    def _get_initial_info(self, url, params=None):
        if params is None:
            params = {}

        max_attempts = params.get('max_attempts', 1)
        for attempt_number in attempts(max_attempts):
            try:
                response = self._session_get(url)
                html = response.text
                yt = regex_search(html, self._YT_INITIAL_DATA_RE)
                yt_initial_data = try_parse_json(yt)

                if response.status_code != 200:
                    # Check for errors
                    title = get_title_of_webpage(html)
                    if response.status_code == 404:
                        raise VideoNotFound(title)
                    elif response.status_code // 100 == 5:  # Server error, retry
                        self.retry(attempt_number, text=title, **params)
                        continue

                if not yt_initial_data:  # Fatal error
                    log('debug', html)
                    raise ParsingError(f'Unable to parse initial video data')

                cfg = regex_search(html, self._YT_CFG_RE)
                ytcfg = try_parse_json(cfg, {})

                player_response = regex_search(
                    html, self._YT_INITIAL_PLAYER_RESPONSE_RE)

                player_response_info = try_parse_json(player_response, {})

                return yt_initial_data, ytcfg, player_response_info

            except RequestException as e:
                self.retry(attempt_number, error=e, **params)

        return None, None, None

    def get_video_data(self, video_id, params=None):
        return self._parse_video_data(video_id, params)[0]

    def _parse_video_data(self, video_id, params=None, video_type='video'):
        details = {}

        if video_type == 'clip':
            original_url = self._YT_CLIP_TEMPLATE.format(video_id)
        else:  # video_type == 'video'
            original_url = self._YT_VIDEO_TEMPLATE.format(video_id)

        yt_initial_data, ytcfg, player_response_info = self._get_initial_info(
            original_url, params)

        if not player_response_info:
            log('debug', yt_initial_data)
            log('warning', f'Unable to parse player response, proceeding with caution')

        streaming_data = player_response_info.get('streamingData') or {}
        first_format = multi_get(streaming_data, 'adaptiveFormats', 0) or multi_get(
            streaming_data, 'formats', 0) or {}

        # Live streaming details
        player_renderer = multi_get(
            player_response_info, 'microformat', 'playerMicroformatRenderer') or {}
        live_details = player_renderer.get('liveBroadcastDetails') or {}

        # Video info
        video_details = player_response_info.get('videoDetails') or {}
        details['title'] = video_details.get('title')
        details['author'] = video_details.get('author')
        details['author_id'] = video_details.get('channelId')
        details['original_video_id'] = video_details.get('videoId')

        # Clip info
        clip_details = player_response_info.get('clipConfig')
        if clip_details:
            details['clip_start_time'] = (float_or_none(
                clip_details.get('startTimeMs', 0)) / 1e3)
            details['clip_end_time'] = (float_or_none(
                clip_details.get('endTimeMs', 0)) / 1e3)
            details['video_type'] = 'clip'

        elif not video_details.get('isLiveContent'):
            details['video_type'] = 'premiere'

        else:
            details['video_type'] = 'video'

        start_timestamp = live_details.get('startTimestamp')
        end_timestamp = live_details.get('endTimestamp')
        details['start_time'] = parse_iso8601(
            start_timestamp) if start_timestamp else None
        details['end_time'] = parse_iso8601(
            end_timestamp) if end_timestamp else None

        details['duration'] = (float_or_none(first_format.get('approxDurationMs', 0)) / 1e3) or float_or_none(
            video_details.get('lengthSeconds')) or float_or_none(player_renderer.get('lengthSeconds'))

        if not details['duration'] and details['start_time'] and details['end_time']:
            details['duration'] = (
                details['end_time'] - details['start_time'])/1e6

        # Parse continuation info
        sub_menu_items = multi_get(yt_initial_data, 'contents', 'twoColumnWatchNextResults', 'conversationBar', 'liveChatRenderer',
                                   'header', 'liveChatHeaderRenderer', 'viewSelector', 'sortFilterSubMenuRenderer', 'subMenuItems') or {}
        details['continuation_info'] = {
            x['title']: x['continuation']['reloadContinuationData']['continuation']
            for x in sub_menu_items
        }

        # live, upcoming or past
        if video_details.get('isLive') or live_details.get('isLiveNow'):
            details['status'] = 'live'

        elif video_details.get('isUpcoming'):
            details['status'] = 'upcoming'

        else:
            details['status'] = 'past'

        return details, player_response_info, yt_initial_data, ytcfg

    def _get_initial_video_info(self, video_id, params=None, video_type='video'):
        """ Get initial YouTube video information. """

        details, player_response_info, yt_initial_data, ytcfg = self._parse_video_data(
            video_id, params, video_type)

        # Error checking
        if not details['continuation_info']:
            # Only raise an error if there is no continuation info. Sometimes you
            # are able to view chat, but not the video (e.g. for very long livestreams)
            playability_status = player_response_info.get(
                'playabilityStatus') or {}
            error_screen = playability_status.get('errorScreen')
            if error_screen:  # There is a error screen visible
                error_reasons = {
                    'reason': '',
                    'subreason': '',
                }
                error_info = try_get_first_value(error_screen)

                for error_reason in error_reasons:
                    text = error_info.get(error_reason) or {}

                    error_reasons[error_reason] = text.get('simpleText') or self._parse_runs(
                        text, False)['message'] or error_info.pop(
                        'itemTitle', '') or error_info.pop(
                            'offerDescription', '') or playability_status.get(error_reason) or ''

                error_message = ''
                for error_reason in error_reasons:
                    if error_reasons[error_reason]:
                        if isinstance(error_reasons[error_reason], str):
                            error_message += f" {error_reasons[error_reason].rstrip('.')}."
                        else:
                            error_message += str(error_reasons[error_reason])

                error_message = error_message.strip()

                status = playability_status.get('status')

                if status == 'ERROR':
                    raise VideoUnavailable(error_message)
                elif status == 'LOGIN_REQUIRED':
                    raise LoginRequired(error_message)
                elif status == 'UNPLAYABLE':
                    raise VideoUnplayable(error_message)
                else:
                    log('debug',
                        f'Unknown playability status: {status}. {playability_status}')
                    error_message = f'{status}: {error_message}'
                    raise VideoUnavailable(error_message)

            # Check for pop up
            popup_info = multi_get(yt_initial_data, 'onResponseReceivedActions',
                                   0, 'openPopupAction', 'popup', 'confirmDialogRenderer')
            if popup_info:
                error_message = multi_get(popup_info, 'title', 'simpleText')
                dialog_messages = multi_get(popup_info, 'dialogMessages') or []
                error_message += '. ' + \
                    ' '.join(map(lambda x: x.get('simpleText'), dialog_messages))
                raise VideoUnavailable(error_message)
            elif not yt_initial_data.get('contents'):
                log('debug', f'Initial YouTube data: {yt_initial_data}')
                raise VideoUnavailable(
                    'Unable to find initial video contents.')
            else:
                # Video exists, but you cannot view chat for some reason
                error_runs = multi_get(yt_initial_data, 'contents', 'twoColumnWatchNextResults', 'conversationBar',
                                       'conversationBarRenderer', 'availabilityMessage', 'messageRenderer', 'text')
                error_message = self._parse_runs(error_runs, False)[
                    'message'] if error_runs else 'Video does not have a chat replay.'

                # Live chat replay was turned off for this video. -> NoChatReplay
                if 'disabled' in error_message:
                    raise ChatDisabled(error_message)
                else:
                    raise NoChatReplay(error_message)

        return details, ytcfg

    def _extract_account_syncid(self, ytcfg):
        sync_ids = ytcfg.get('DATASYNC_ID').split('||')
        if len(sync_ids) >= 2 and sync_ids[1]:
            # datasyncid is of the form "channel_syncid||user_syncid" for secondary channel
            # and just "user_syncid||" for primary channel. We only want the channel_syncid
            return sync_ids[0]

        # ytcfg includes channel_syncid if on secondary channel
        return ytcfg.get('DELEGATED_SESSION_ID')

    def _generate_headers(self, ytcfg):
        headers = {
            'origin': self._YT_HOME,
            'x-youtube-client-name': str(ytcfg.get('INNERTUBE_CONTEXT_CLIENT_NAME')),
            'x-youtube-client-version': str(ytcfg.get('INNERTUBE_CLIENT_VERSION')),
            'x-origin': self._YT_HOME,
            'x-goog-authuser': '0'
        }

        identity_token = ytcfg.get('ID_TOKEN')
        if identity_token:
            headers['x-youtube-identity-token'] = identity_token

        account_syncid = self._extract_account_syncid(ytcfg)
        if account_syncid:
            headers['x-goog-pageid'] = account_syncid

        session_index = ytcfg.get('SESSION_INDEX')
        if account_syncid or session_index:
            headers['x-goog-authuser'] = session_index or 0

        visitor_data = multi_get(
            ytcfg, 'INNERTUBE_CONTEXT', 'client', 'visitorData')
        if visitor_data:
            headers['x-goog-visitor-id'] = visitor_data

        auth = self._generate_sapisidhash_header()
        if auth:
            headers['authorization'] = auth

        return headers

    def _get_chat_messages(self, initial_info, ytcfg, params):

        initial_continuation_info = initial_info.get('continuation_info') or {}
        if len(initial_continuation_info) < 2:
            raise NoContinuation(
                f'Initial continuation information could not be found: {initial_info}')

        status = initial_info.get('status')

        # duration = initial_info.get('duration')
        # stream_start_time = initial_info.get('start_time')
        offset = initial_info.get('offset')  # Clips

        start_time = ensure_seconds(params.get('start_time'))
        end_time = ensure_seconds(params.get('end_time'))

        # Top chat replay - Some messages, such as potential spam, may not be visible
        # Live chat replay - All messages are visible
        chat_type = params.get('chat_type').title()  # Live or Top
        continuation_index = 0 if chat_type == 'Top' else 1
        continuation_info = list(initial_continuation_info.items())[
            continuation_index]
        continuation = continuation_info[1]
        log('debug', f'Getting {chat_type} chat ({continuation_info[0]}).')

        is_replay = status == 'past'

        api_type = 'live_chat'
        if is_replay:
            api_type += '_replay'

        init_page = self._YOUTUBE_INIT_API_TEMPLATE.format(
            api_type, continuation)

        api_key = ytcfg.get('INNERTUBE_API_KEY')

        continuation_url = self._YOUTUBE_CHAT_API_TEMPLATE.format(
            api_type, api_key)
        offset_milliseconds = (
            start_time * 1000) if isinstance(start_time, (float, int)) else None

        # force_no_timeout = params.get('force_no_timeout')

        max_attempts = params.get('max_attempts')

        messages_groups_to_add = params.get('message_groups') or []
        messages_types_to_add = params.get('message_types') or []

        invalid_groups = set(messages_groups_to_add) - \
            self._MESSAGE_GROUPS.keys()
        if 'all' not in messages_groups_to_add and invalid_groups:
            raise InvalidParameter(
                f'Invalid groups specified: {invalid_groups}')

        self.check_for_invalid_types(
            messages_types_to_add, self._MESSAGE_TYPES)

        # Generate base headers and update session headers
        self.update_session_headers(self._generate_headers(ytcfg))

        self.update_session_headers({
            'content-type': 'application/json',
            'referer': init_page
        })

        innertube_context = ytcfg.get('INNERTUBE_CONTEXT') or {}

        message_count = 0
        first_time = True
        click_tracking_params = None

        while True:
            continuation_params = {
                'context': innertube_context,
                'continuation': continuation
            }

            # Update authentication header, if necessary
            auth = self._generate_sapisidhash_header()
            if auth:
                self.update_session_headers({
                    'authorization': auth
                })

            if first_time:
                # must run to get first few messages, otherwise might miss some
                yt_info = self._get_initial_info(init_page, params)[0]

            else:
                if is_replay and offset_milliseconds is not None:
                    continuation_params['currentPlayerState'] = {
                        'playerOffsetMs': offset_milliseconds}

                if click_tracking_params:
                    continuation_params['context']['clickTracking'] = {
                        'clickTrackingParams': click_tracking_params}

                yt_info = self._get_continuation_info(
                    continuation_url, params, json=continuation_params)

            debug_info = {
                'click_tracking': multi_get(continuation_params, 'context', 'clickTracking'),
                'continuation': multi_get(continuation_params, 'continuation')
            }
            log('debug', [
                f'Continuation parameters: {debug_info}',
                f"Session headers: {', '.join(self.session.headers.keys())}"
            ])

            logged_in_info = multi_get(
                yt_info, 'responseContext', 'serviceTrackingParams', 1, 'params', 0)
            log('debug', f'Logged-in info: {logged_in_info}')

            info = multi_get(yt_info, 'continuationContents',
                             'liveChatContinuation')
            if not info:
                log('debug', f'No continuation information found: {yt_info}')
                return

            actions = info.get('actions') or []

            if actions:
                for action in actions:
                    data = {}

                    # if it is a replay chat item action, must re-base it
                    replay_chat_item_action = action.get(
                        'replayChatItemAction')
                    if replay_chat_item_action:
                        offset_time = replay_chat_item_action.get(
                            'videoOffsetTimeMsec')
                        if offset_time:
                            data['time_in_seconds'] = float(offset_time) / 1000

                        action = replay_chat_item_action['actions'][0]

                    action.pop('clickTrackingParams', None)
                    original_action_type = try_get_first_key(action)

                    data['action_type'] = camel_case_split(
                        remove_suffixes(original_action_type, ('Action', 'Command')))

                    original_message_type = None
                    original_item = {}

                    # We now parse the info and get the message
                    # type based on the type of action
                    if original_action_type in self._KNOWN_ITEM_ACTION_TYPES:
                        original_item = multi_get(
                            action, original_action_type, 'item')

                        original_message_type = try_get_first_key(
                            original_item)
                        data = self._parse_item(original_item, data, offset)

                    elif original_action_type in self._KNOWN_REMOVE_ACTION_TYPES:
                        original_item = action
                        if original_action_type == 'markChatItemAsDeletedAction':
                            original_message_type = 'deletedMessage'
                        else:  # markChatItemsByAuthorAsDeletedAction
                            original_message_type = 'banUser'

                        data = self._parse_item(original_item, data, offset)

                    elif original_action_type in self._KNOWN_REPLACE_ACTION_TYPES:
                        original_item = multi_get(
                            action, original_action_type, 'replacementItem')

                        original_message_type = try_get_first_key(
                            original_item)
                        data = self._parse_item(original_item, data, offset)

                    elif original_action_type in self._KNOWN_TOOLTIP_ACTION_TYPES:
                        original_item = multi_get(
                            action, original_action_type, 'tooltip')

                        original_message_type = try_get_first_key(
                            original_item)
                        data = self._parse_item(original_item, data, offset)

                    elif original_action_type in self._KNOWN_ADD_BANNER_TYPES:
                        original_item = multi_get(
                            action, original_action_type, 'bannerRenderer')

                        if original_item:
                            original_message_type = try_get_first_key(
                                original_item)

                            header = original_item[original_message_type].get(
                                'header')
                            parsed_header = self._parse_item(
                                header, offset=offset)
                            header_message = parsed_header.get('message')

                            contents = original_item[original_message_type].get(
                                'contents')
                            parsed_contents = self._parse_item(
                                contents, offset=offset)

                            data.update(parsed_header)
                            data.update(parsed_contents)
                            data['header_message'] = header_message
                        else:
                            debug_log(
                                'No bannerRenderer item',
                                f'Action type: {original_action_type}',
                                f'Action: {action}',
                                f'Parsed data: {data}'
                            )

                    elif original_action_type in self._KNOWN_REMOVE_BANNER_TYPES:
                        original_item = action
                        original_message_type = 'removeBanner'
                        data = self._parse_item(original_item, data, offset)

                    elif original_action_type in self._KNOWN_IGNORE_ACTION_TYPES:
                        continue  # ignore these

                    else:
                        # not processing these
                        debug_log(
                            f'Unknown action: {original_action_type}',
                            action,
                            data
                        )

                    test_for_missing_keys = original_item.get(
                        original_message_type, {}).keys()
                    missing_keys = test_for_missing_keys - self._KNOWN_KEYS

                    if not data:
                        debug_log(
                            f'Parse of action returned empty results: {original_action_type}',
                            action
                        )

                    if missing_keys:
                        debug_log(
                            f'Missing keys found: {missing_keys}',
                            f'Message type: {original_message_type}',
                            f'Action type: {original_action_type}',
                            f'Action: {action}',
                            f'Parsed data: {data}'
                        )

                    if original_message_type:

                        new_index = remove_prefixes(
                            original_message_type, 'liveChat')
                        new_index = remove_suffixes(new_index, 'Renderer')
                        data['message_type'] = camel_case_split(new_index)

                        # TODO add option to keep placeholder items
                        if original_message_type in self._KNOWN_IGNORE_MESSAGE_TYPES:
                            continue
                            # skip placeholder items
                        elif original_message_type not in self._KNOWN_ACTION_TYPES[original_action_type]:
                            debug_log(
                                f'Unknown message type "{original_message_type}" for action "{original_action_type}"',
                                f"New message type: {data['message_type']}",
                                f'Action: {action}',
                                f'Parsed data: {data}'
                            )

                    else:  # no type # can ignore message
                        debug_log(
                            'No message type',
                            f'Action type: {original_action_type}',
                            f'Action: {action}',
                            f'Parsed data: {data}'
                        )
                        continue

                    # check whether to skip this message or not, based on its type

                    to_add = self._must_add_item(
                        data,
                        self._MESSAGE_GROUPS,
                        messages_groups_to_add,
                        messages_types_to_add
                    )

                    if not to_add:
                        continue

                    # if from a replay, check whether to skip this message or not, based on its time
                    if is_replay:
                        # assume message is at beginning if it does not have a time component
                        time_in_seconds = data.get(
                            'time_in_seconds', 0) + (offset or 0)

                        before_start = start_time is not None and time_in_seconds < start_time
                        after_end = end_time is not None and time_in_seconds > end_time

                        if first_time and before_start:
                            continue  # first time and invalid start time
                        elif before_start or after_end:
                            return  # while actually searching, if time is invalid

                    # try to reconstruct time in seconds from timestamp and stream start
                    # if data.get('time_in_seconds') is None and data.get('timestamp') and stream_start_time:
                    #     data['time_in_seconds'] = (data['timestamp'] - stream_start_time)/1e6
                    #     data['time_text'] = seconds_to_time(int(data['time_in_seconds']))

                    message_count += 1
                    yield data

                log('debug', f'Total number of messages: {message_count}')
            elif is_replay:
                # no more actions to process in a chat replay
                break
            else:
                # otherwise, is live, so keep trying
                log('debug', 'No actions to process.')

            # assume there are no more chat continuations
            no_continuation = True

            # parse the continuation information
            for cont in info.get('continuations') or []:

                continuation_key = try_get_first_key(cont)
                continuation_info = cont[continuation_key]

                log('debug', f'Continuation info: {continuation_info}')

                if continuation_key in self._KNOWN_CHAT_CONTINUATIONS:

                    # set new chat continuation
                    # overwrite if there is continuation data
                    continuation = continuation_info.get('continuation')

                    click_tracking_params = continuation_info.get(
                        'clickTrackingParams') or continuation_info.get('trackingParams')
                    # there is a chat continuation
                    no_continuation = False

                elif continuation_key in self._KNOWN_SEEK_CONTINUATIONS:
                    pass
                    # ignore these continuations

                else:
                    debug_log(
                        f'Unknown continuation: {continuation_key}',
                        cont
                    )

                # sometimes continuation contains timeout info
                sleep_duration = continuation_info.get('timeoutMs')
                # and not actions:# and not force_no_timeout:
                if sleep_duration:
                    # Timeouts help prevent 429 errors (caused by too many requests).
                    #
                    # A single request to the YouTube live chat endpoint seems to only
                    # go back around 10 seconds (only retrieving around 150-200 messages
                    # at any given time).
                    #
                    # For very large livestreams, YouTube sometimes sets timeouts to be
                    # more than 10 seconds (most likely to alleviate server stress).
                    # This means that, normally, users will not be able to see all chat
                    # messages (leaving a gap of timeout - 10 seconds).
                    #
                    # To get around this, we clamp the timeout to be between 0 and 8000
                    # milliseconds (a 2 second window for making the next request).
                    # This ensures that no messages are missed and we do spam YouTube
                    # with requests (which may lead to 429 errors or IP blocking).

                    sleep_duration = max(min(sleep_duration, 8000), 0)

                    log('debug', f'Sleeping for {sleep_duration}ms.')
                    interruptible_sleep(sleep_duration / 1000)

            if no_continuation:  # no continuation, end
                break

            if first_time:
                first_time = False

    def _get_chat_by_clip_id(self, match, params):
        return self.get_chat_by_clip_id(match.group('id'), params)

    def get_chat_by_clip_id(self, clip_id, params):

        initial_info, ytcfg = self._get_initial_video_info(
            clip_id, params, video_type='clip')

        initial_info['offset'] = clip_start_time = initial_info.get(
            'clip_start_time')
        clip_end_time = initial_info.get('clip_end_time')

        max_duration = clip_end_time - clip_start_time

        params['start_time'] = ensure_seconds(
            params.get('start_time'), 0) + clip_start_time
        params['end_time'] = ensure_seconds(params.get(
            'end_time'), max_duration) + clip_start_time

        return Chat(
            self._get_chat_messages(initial_info, ytcfg, params),
            id=clip_id,
            **initial_info
        )

    def _get_chat_by_user(self, match, params):
        match_id = match.group('id')
        user_type = match.group('type')  # channel|c|user

        if user_type == 'channel':
            return self.get_chat_by_channel_id(match_id, params)

        if user_type == 'user':
            return self.get_chat_by_user_id(match_id, params)

        # Otherwise assume custom username
        return self.get_chat_by_custom_username(match_id, params)

    def get_chat_by_channel_id(self, channel_id, params):
        return self._get_chat_by_user_args({
            'channel_id': channel_id
        }, params)

    def get_chat_by_user_id(self, user_id, params):
        """
        Such as NASAtelevision in https://www.youtube.com/user/NASAtelevision

        :param user_id:
        :type user_id: [type]
        """
        return self._get_chat_by_user_args({
            'user_id': user_id
        }, params)

    def get_chat_by_custom_username(self, custom_username, params):
        return self._get_chat_by_user_args({
            'custom_username': custom_username
        }, params)

    def _get_chat_by_user_args(self, user_video_args, params):
        # TODO add param for wait time
        # params['exit_on_fail'] = True

        title = try_get_first_value(user_video_args)
        chat_item = Chat(title=title, id=title)  # Create empty chat object
        chat_item.chat = self._get_chat_messages_by_user_args(
            user_video_args, chat_item, params)

        return chat_item

    def _get_chat_messages_by_user_args(self, user_video_args, chat_item, params):
        # chat_item allows to change title and info based on new info

        list_of_vids_to_ignore = params.get('ignore') or []

        sleep_amount = 30  # params.get('retry_timeout')

        while True:
            for video_status in ('live', 'upcoming'):
                # prioritise live videos
                for video in self.get_user_videos(**user_video_args, video_status=video_status, params=params):
                    video_id = video['video_id']
                    video_title = video['title']

                    if video_id in list_of_vids_to_ignore:
                        log('debug', f'Skipping video with ID: "{video_id}"')
                        continue

                    try:
                        chat = self.get_chat_by_video_id(video_id, params)

                        log('info',
                            f"Found a{'n upcoming' if video_status == 'upcoming' else ''} livestream: \"{video_title}\" ({video_id}).")

                        for key, value in vars(chat).items():  # Update chat item
                            if key != 'chat' and not key.startswith('_'):
                                setattr(chat_item, key, value)

                        yield from chat
                        break

                    except ChatDownloaderError as e:
                        # For some reason, doesn't work
                        log('warning',
                            f"Unable to get chat for \"{video['title']}\" ({video_id}) due to an error: \"{e}\"")

            log('info',
                f'There are no active or upcoming livestreams with a live chat. Retrying in {sleep_amount} seconds.')
            interruptible_sleep(sleep_amount)

            # continue forever, until reaching a video with valid chat

    def get_chat_by_video_id(self, video_id, params):
        """Get chat messages for a YouTube video, given its ID.

        :param video_id: YouTube video ID
        :type video_id: str
        :return: Chat object for the corresponding YouTube video
        :rtype: Chat
        """
        initial_info, ytcfg = self._get_initial_video_info(video_id, params)

        return Chat(
            self._get_chat_messages(initial_info, ytcfg, params),
            id=video_id,
            **initial_info
        )

    def _get_chat_by_video_id(self, match, params):
        return self.get_chat_by_video_id(match.group('id'), params)



#!/usr/bin/env python

"""The setup script."""
# See https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/#setup-args

from setuptools import setup, find_packages

# Get metadata without importing the package
with open('chat_analyzer/metadata.py') as metadata_file:
    exec(metadata_file.read())
    metadata = locals()

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [ # TODO: Check if this is correct
    # For the chat_analyzer
    'argparse',
    'numpy',

    # For included chat_downloader
    'requests',
    'isodate',
    # 'argparse',
    'docstring-parser',
    'colorlog',
    'websocket-client'
]

setup(
    name=metadata['__title__'],
    version = metadata['__version__'],
    description=metadata['__summary__'],
    long_description=readme,
    long_description_content_type='text/x-rst',
    url=metadata['__url__'],
    author=metadata['__author__'],
    author_email=metadata['__email__'],
    license=metadata['__license__'],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        'Topic :: Communications',
        'Topic :: Communications :: Chat',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3, or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='chat analytics analyze message twitch youtube streamer video VOD editor',
    project_urls={
        'Website': metadata['__website__'],
        'Source': metadata['__url__'],
        # TODO Add Documentation, Funding/Say thanks
    },
    packages=find_packages(include=['chat_analyzer', 'chat_analyzer.*']),
    install_requires=requirements,
    python_requires='>=3',
    # TODO wtf is this? do we need to include the stuff from the chat_downloader as well?
    # package_data={
    #     'chat_downloader': ['formatting/*.json']
    # },
    entry_points={
        'console_scripts': [
            'chat_analyzer=chat_analyzer.cli:main',
            'chat-analyzer=chat_analyzer.cli:main',
        ],
    },

    # TODO: Reference Xenova's setup for the chat_downloader to see if theres anything worth adding later
)



"""Debugging module for chat_downloader"""

import sys
import os

from .metadata import __name__ as logger_name
from .utils.core import pause

from enum import Enum


class TestingException(Exception):
    """Raised when something unexpected happens while in testing mode"""


class TestingModes(Enum):
    # Currently unused
    EXIT_ON_ERROR = 4
    PAUSE_ON_ERROR = 3

    # In use
    EXIT_ON_DEBUG = 2
    PAUSE_ON_DEBUG = 1
    NONE = 0


TESTING_MODE = TestingModes.NONE


def set_testing_mode(new_mode):
    global TESTING_MODE
    TESTING_MODE = new_mode


def log(level, items, to_pause=False, to_exit=False):
    logger_at_level = getattr(logger, level, None)
    if logger_at_level:
        if not isinstance(items, (tuple, list)):
            items = [items]
        for item in items:
            logger_at_level(item)

        if to_exit and TESTING_MODE in (TestingModes.EXIT_ON_ERROR, TestingModes.EXIT_ON_DEBUG):
            raise TestingException(
                'Testing exception encountered, exiting program')

        if to_pause and TESTING_MODE in (TestingModes.PAUSE_ON_ERROR, TestingModes.PAUSE_ON_DEBUG):
            pause()


def debug_log(*items):
    """Method which simplifies the logging of debugging messages"""
    log('debug', items, True, True)


try:
    import colorama
    colorama.init()
except (ImportError, OSError):
    HAS_COLORAMA = False
else:
    HAS_COLORAMA = True


def supports_colour():
    """
    Return True if the running system's terminal supports colour,
    and False otherwise.

    Adapted from https://github.com/django/django/blob/master/django/core/management/color.py
    """
    def vt_codes_enabled_in_windows_registry():
        """
        Check the Windows Registry to see if VT code handling has been enabled
        by default, see https://superuser.com/a/1300251/447564.
        """
        try:
            # winreg is only available on Windows.
            import winreg
        except ImportError:
            return False
        else:
            reg_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, 'Console')
            try:
                reg_key_value, _ = winreg.QueryValueEx(
                    reg_key, 'VirtualTerminalLevel')
            except FileNotFoundError:
                return False
            else:
                return reg_key_value == 1

    # isatty is not always implemented, #6223.
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()

    return is_a_tty and (
        sys.platform != 'win32' or
        HAS_COLORAMA or
        'ANSICON' in os.environ or

        # Windows Terminal supports VT codes.
        'WT_SESSION' in os.environ or

        # Microsoft Visual Studio Code's built-in terminal supports colors.
        os.environ.get('TERM_PROGRAM') == 'vscode' or
        vt_codes_enabled_in_windows_registry()
    )


if supports_colour():
    import colorlog as log_module
    handler = log_module.StreamHandler()
    handler.setFormatter(log_module.ColoredFormatter(
        '[%(log_color)s%(levelname)s%(reset)s] %(message)s',
        log_colors={
            'DEBUG': 'cyan',
            'INFO': 'green',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'bold_red',
        })
    )

else:  # fallback support
    import logging as log_module
    handler = log_module.StreamHandler()
    handler.setFormatter(log_module.Formatter('[%(levelname)s] %(message)s'))

# Create logger object for this module
logger = log_module.getLogger(logger_name)

# Define which loggers to display
loggers = [log_module.getLogger(name) for name in (logger_name, 'urllib3')]
for logger in loggers:
    logger.addHandler(handler)


def set_log_level(level):
    level_name = level.upper()
    for logger in loggers:
        logger.setLevel(level_name)


def disable_logger():
    logger.disabled = True



import inspect
import datetime
import re
import sys
import locale
import collections.abc
import io
import json
import base64


def base64_encode(text):
    return base64.b64encode(text.encode()).decode()


def timestamp_to_microseconds(timestamp):
    """Convert RFC3339 timestamp to microseconds. This is needed since
        ``datetime.datetime.strptime()`` does not support nanosecond precision.

    :param timestamp: RFC3339 timestamp
    :type timestamp: str
    :return: The number of microseconds of the timestamp
    :rtype: int
    """

    info = list(filter(None, re.split(r'[\.|Z]{1}', timestamp))) + [0]
    return round((datetime.datetime.strptime(f'{info[0]}Z', '%Y-%m-%dT%H:%M:%SZ').timestamp() + float(f'0.{info[1]}')) * 1e6)


def time_to_seconds(time):
    """Convert timestamp string of the form 'hh:mm:ss' to seconds.

    :param time: Timestamp of the form 'hh:mm:ss'
    :type time: str
    :return: The corresponding number of seconds
    :rtype: int
    """
    if not time:
        return 0
    return int(sum(abs(int(x)) * 60 ** i for i, x in enumerate(reversed(time.replace(',', '').split(':')))) * (-1 if time[0] == '-' else 1))


def seconds_to_time(seconds, format='{}:{:02}:{:02}', remove_leading_zeroes=True):
    """Convert seconds to timestamp.

    :param seconds: Number of seconds
    :type seconds: int
    :param format: The format string with elements representing hours, minutes and seconds. Defaults to '{}:{:02}:{:02}'
    :type format: str, optional
    :param remove_leading_zeroes: Whether to remove leading zeroes when seconds > 60, defaults to True
    :type remove_leading_zeroes: bool, optional
    :return: The corresponding timestamp string
    :rtype: str
    """
    h, remainder = divmod(abs(int(seconds)), 3600)
    m, s = divmod(remainder, 60)
    time_string = format.format(h, m, s)
    return ('-' if seconds < 0 else '') + (re.sub(r'^0:0?', '', time_string) if remove_leading_zeroes else time_string)


def microseconds_to_timestamp(microseconds, format='%Y-%m-%d %H:%M:%S'):
    """Convert unix time to human-readable timestamp.

    :param microseconds: UNIX microseconds
    :type microseconds: float
    :param format: The format string, defaults to '%Y-%m-%d %H:%M:%S'. For
        information on supported codes, see https://strftime.org/ and
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    :type format: str, optional
    :return: Human readable timestamp corresponding to the format
    :rtype: str
    """
    return datetime.datetime.fromtimestamp(microseconds // 1000000).strftime(format)


def ensure_seconds(time, default=None):
    """Ensure time is returned in seconds.

    :param time: The time, in seconds or 'hh:mm:ss'.
    :type time: Union[float, str]
    :param default: Returns this if unable to parse the time, defaults to None
    :type default: object, optional
    :return: The corresponding number of seconds
    :rtype: float
    """
    if time is None:
        return default

    try:
        return float(time)
    except ValueError:
        return time_to_seconds(time)
    except Exception:
        return default


def arbg_int_to_rgba(argb_int):
    """Convert ARGB integer to RGBA array.

    :param argb_int: ARGB integer
    :type argb_int: int
    :return: RGBA array
    :rtype: list[int]
    """
    red = (argb_int >> 16) & 255
    green = (argb_int >> 8) & 255
    blue = argb_int & 255
    alpha = (argb_int >> 24) & 255
    return [red, green, blue, alpha]


def rgba_to_hex(colours):
    """Convert RGBA array to hex colour.

    :param colours: RGBA array
    :type colours: list[int]
    :return: Corresponding hexadecimal representation
    :rtype: str
    """
    return '#{:02x}{:02x}{:02x}{:02x}'.format(*colours)


def regex_search(text, pattern, group=1, default=None):
    match = re.search(pattern, text)
    return match.group(group) if match else default


def get_title_of_webpage(html):
    return regex_search(html, '<title(?:[^>]*)>(.*?)</title>')


def int_or_none(v, default=None):
    try:
        return int(v)
    except (ValueError, TypeError):
        return default


def float_or_none(v, default=None):
    try:
        return float(v)
    except (ValueError, TypeError):
        return default


def str_or_none(v, default=None):
    try:
        return str(v)
    except (ValueError, TypeError):
        return default


def try_get_first_key(dictionary, default=None):
    try:
        return next(iter(dictionary))
    except Exception:
        return default


def try_get_first_value(dictionary, default=None):
    try:
        return next(iter(dictionary.values()))
    except Exception:
        return default


def try_parse_json(text, default=None):
    try:
        return json.loads(text)
    except (json.decoder.JSONDecodeError, TypeError):
        return default


def wrap_as_list(item):
    """Wraps an item in a list, if it is not already iterable

    :param item: The item to wrap
    :type item: object
    :return: The wrapped item
    :rtype: Union[list, tuple]
    """
    if not isinstance(item, (list, tuple)):
        item = [item]
    return item


def remove_prefixes(text, prefixes):
    for prefix in wrap_as_list(prefixes):
        if text.startswith(prefix):
            text = text[len(prefix):]
    return text


def remove_suffixes(text, suffixes):
    for suffix in wrap_as_list(suffixes):
        if text.endswith(suffix):
            text = text[0:-len(suffix):]
    return text


def update_dict_without_overwrite(original, new):
    original.update({key: new[key] for key in new if key not in original})
    return original


def camel_case_split(word):
    return '_'.join(re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)', word)).lower()


def replace_with_underscores(text, sep='-'):
    return text.replace(sep, '_')


def multi_get(dictionary, *keys, default=None):
    current = dictionary
    for key in keys:
        if isinstance(current, dict):
            current = current.get(key, default)
        elif isinstance(current, (list, tuple)) and isinstance(key, int):
            try:
                current = current[key]
            except IndexError:
                return default
        else:
            return default
    return current


def flatten_json(original_json):
    final = {}

    def flatten(item, prefix=''):
        if isinstance(item, dict):
            for key in item:
                flatten(item[key], f'{prefix}{key}.')
        elif isinstance(item, list):
            for index in range(len(item)):
                flatten(item[index], f'{prefix}{index}.')
        else:
            final[prefix[:-1]] = item
    flatten(original_json)

    return final


def attempts(max_attempts):
    return range(1, max_attempts + 1)


def preferredencoding():
    """Get preferred encoding.
    Returns the best encoding scheme for the system, based on
    locale.getpreferredencoding() and some further tweaks.
    """
    try:
        pref = locale.getpreferredencoding()
        'TEST'.encode(pref)
    except Exception:
        pref = 'utf-8'

    return pref


def _windows_write_string(s, out, skip_errors=True):
    """ Returns True if the string was written using special methods,
    False if it has yet to be written out."""
    # Adapted from http://stackoverflow.com/a/3259271/35070

    import ctypes
    import ctypes.wintypes

    WIN_OUTPUT_IDS = {
        1: -11,
        2: -12,
    }

    try:
        fileno = out.fileno()
    except AttributeError:
        # If the output stream doesn't have a fileno, it's virtual
        return False
    except io.UnsupportedOperation:
        # Some strange Windows pseudo files?
        return False
    if fileno not in WIN_OUTPUT_IDS:
        return False

    GetStdHandle = ctypes.WINFUNCTYPE(
        ctypes.wintypes.HANDLE, ctypes.wintypes.DWORD)(
        ('GetStdHandle', ctypes.windll.kernel32))
    h = GetStdHandle(WIN_OUTPUT_IDS[fileno])

    WriteConsoleW = ctypes.WINFUNCTYPE(
        ctypes.wintypes.BOOL, ctypes.wintypes.HANDLE, ctypes.wintypes.LPWSTR,
        ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD),
        ctypes.wintypes.LPVOID)(('WriteConsoleW', ctypes.windll.kernel32))
    written = ctypes.wintypes.DWORD(0)

    GetFileType = ctypes.WINFUNCTYPE(ctypes.wintypes.DWORD, ctypes.wintypes.DWORD)(
        ('GetFileType', ctypes.windll.kernel32))
    FILE_TYPE_CHAR = 0x0002
    FILE_TYPE_REMOTE = 0x8000
    GetConsoleMode = ctypes.WINFUNCTYPE(
        ctypes.wintypes.BOOL, ctypes.wintypes.HANDLE,
        ctypes.POINTER(ctypes.wintypes.DWORD))(
        ('GetConsoleMode', ctypes.windll.kernel32))
    INVALID_HANDLE_VALUE = ctypes.wintypes.DWORD(-1).value

    def not_a_console(handle):
        if handle == INVALID_HANDLE_VALUE or handle is None:
            return True
        return ((GetFileType(handle) & ~FILE_TYPE_REMOTE) != FILE_TYPE_CHAR or GetConsoleMode(handle, ctypes.byref(ctypes.wintypes.DWORD())) == 0)

    if not_a_console(h):
        return False

    def next_nonbmp_pos(s):
        try:
            return next(i for i, c in enumerate(s) if ord(c) > 0xffff)
        except StopIteration:
            return len(s)

    while s:
        count = min(next_nonbmp_pos(s), 1024)

        ret = WriteConsoleW(
            h, s, count if count else 2, ctypes.byref(written), None)
        if ret == 0:
            if skip_errors:
                continue
            else:
                raise OSError('Failed to write string')
        if not count:  # We just wrote a non-BMP character
            assert written.value == 2
            s = s[1:]
        else:
            assert written.value > 0
            s = s[written.value:]
    return True


def safe_print(*objects, sep=' ', end='\n', out=None, encoding=None, flush=False):
    """
    Ensure printing to standard output can be done safely (especially on Windows).
    There are usually issues with printing emojis and non utf-8 characters.
    """

    output_string = sep.join(map(lambda x: str(x), objects)) + end

    if out is None:
        out = sys.stdout

    if sys.platform == 'win32' and encoding is None and hasattr(out, 'fileno'):
        if _windows_write_string(output_string, out):
            return

    if 'b' in getattr(out, 'mode', '') or not hasattr(out, 'buffer'):
        out.write(output_string)
    else:
        enc = encoding or getattr(out, 'encoding', None) or preferredencoding()
        byt = output_string.encode(enc, 'ignore')
        out.buffer.write(byt)

    if flush and hasattr(out, 'flush'):
        out.flush()


def nested_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            a = d.get(k, {})
            if isinstance(a, dict):
                d[k] = nested_update(a, v)
            else:
                d[k] = v
        else:
            d[k] = v
    return d


def pause(text='Press Enter to continue...'):
    input(text)


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def safe_path(text, replace_char='_'):
    """Ensure generated file name/path is safe
    https://stackoverflow.com/a/31976060
    """
    return re.sub(r'[\/:*?"<>|]', replace_char, text)


# Adapted from https://github.com/micktwomey/pyiso8601/
ISO8601_REGEX = re.compile(
    r"""
    (?P<year>[0-9]{4})
    (
        (
            (-(?P<monthdash>[0-9]{1,2}))
            |
            (?P<month>[0-9]{2})
            (?!$)  # Don't allow YYYYMM
        )
        (
            (
                (-(?P<daydash>[0-9]{1,2}))
                |
                (?P<day>[0-9]{2})
            )
            (
                (
                    (?P<separator>[ T])
                    (?P<hour>[0-9]{2})
                    (:{0,1}(?P<minute>[0-9]{2})){0,1}
                    (
                        :{0,1}(?P<second>[0-9]{1,2})
                        ([.,](?P<second_fraction>[0-9]+)){0,1}
                    ){0,1}
                    (?P<timezone>
                        Z
                        |
                        (
                            (?P<tz_sign>[-+])
                            (?P<tz_hour>[0-9]{2})
                            :{0,1}
                            (?P<tz_minute>[0-9]{2}){0,1}
                        )
                    ){0,1}
                ){0,1}
            )
        ){0,1}  # YYYY-MM
    ){0,1}  # YYYY only
    $
    """,
    re.VERBOSE,
)

UTC = datetime.timezone.utc



def parse_timezone(matches, default_timezone=UTC):
    """Parses ISO 8601 time zone specs into tzinfo offsets"""
    tz = matches.get('timezone', None)
    if tz == 'Z':
        return UTC

    if tz is None:
        return default_timezone
    sign = matches.get('tz_sign', None)
    hours = int(matches.get('tz_hour', 0))
    minutes = int(matches.get('tz_minute', 0))
    description = f'{sign}{hours:02d}:{minutes:02d}'
    if sign == '-':
        hours = -hours
        minutes = -minutes
    return datetime.timezone(datetime.timedelta(hours=hours, minutes=minutes), description)


def parse_date(datestring, default_timezone=UTC):
    """Parses ISO 8601 dates into datetime objects
    The timezone is parsed from the date string. However it is quite common to
    have dates without a timezone (not strictly correct). In this case the
    default timezone specified in default_timezone is used. This is UTC by
    default.
    :param datestring: The date to parse as a string
    :param default_timezone: A datetime tzinfo instance to use when no timezone
                             is specified in the datestring. If this is set to
                             None then a naive datetime object is returned.
    :returns: A datetime.datetime instance
    :raises: ValueError when there is a problem parsing the date or
             constructing the datetime instance.
    """
    try:
        m = ISO8601_REGEX.match(datestring)
    except Exception as e:
        raise ValueError(e)

    if not m:
        raise ValueError(f'Unable to parse date string {datestring!r}')

    groups = {k: v for k, v in m.groupdict().items() if v is not None}

    try:
        return datetime.datetime(
            year=int(groups.get('year', 0)),
            month=int(groups.get('month', groups.get('monthdash', 1))),
            day=int(groups.get('day', groups.get('daydash', 1))),
            hour=int(groups.get('hour', 0)),
            minute=int(groups.get('minute', 0)),
            second=int(groups.get('second', 0)),
            microsecond=int(
                float(f"0.{groups.get('second_fraction', 0)}") * 1e6
            ),
            tzinfo=parse_timezone(groups, default_timezone=default_timezone),
        )
    except Exception as e:
        raise ValueError(e)


def parse_iso8601(data_str):
    return parse_date(data_str).timestamp() * 1e6



from .common import (
    Chat,
    BaseChatDownloader,
    Remapper as r,
    Image
)
from ..utils.core import (
    remove_prefixes,
    multi_get,
    seconds_to_time,
    camel_case_split,
    ensure_seconds,
    attempts,
    regex_search,
    base64_encode,
)
from ..utils.timed_utils import interruptible_sleep

from ..errors import (
    SiteError,
    VideoUnavailable,
    LoginRequired
)

from ..debugging import (log, debug_log)

import json
import re
from json.decoder import JSONDecodeError
from requests.exceptions import RequestException


class FacebookError(SiteError):
    """Raised when an error occurs with a Facebook video."""
    pass


class RateLimitError(FacebookError):
    """Raised when the user has been rate-limited."""
    pass


class FacebookChatDownloader(BaseChatDownloader):
    _FB_HOMEPAGE = 'https://www.facebook.com'

    _INITIAL_DATR_REGEX = r'_js_datr\",\"([^\"]+)'
    _INITIAL_LSD_REGEX = r'<input.*?name=\"lsd\".*?value=\"([^\"]+)[^>]*>'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # update headers for all subsequent FB requests
        self.update_session_headers({
            'Content-Type': 'application/x-www-form-urlencoded',
            'Referer': self._FB_HOMEPAGE,  # Required'
        })

        initial_data = self._session_get(self._FB_HOMEPAGE).text

        datr = regex_search(initial_data, self._INITIAL_DATR_REGEX)
        if not datr:
            raise FacebookError(f'Unable to set datr cookie: {initial_data}')

        self.set_cookie_value('.facebook.com', 'datr', datr)
        self.set_cookie_value('.facebook.com', 'wd', '1920x1080')

        lsd = regex_search(initial_data, self._INITIAL_LSD_REGEX)
        if not lsd:
            raise FacebookError(f'Unable to set lsd cookie: {initial_data}')

        self.lsd = lsd
        self.update_session_headers({
            'x-fb-lsd': lsd,
            'upgrade-insecure-requests': '1',
            'cache-control': 'max-age=0'
        })

    _NAME = 'facebook.com'
    # Regex provided by youtube-dl

    _VALID_URLS = {
        '_get_chat_by_video_id': r'''(?x)
            (?:
                https?://
                    (?:[\w-]+\.)?(?:facebook\.com)/
                    (?:[^#]*?\#!/)?
                    (?:[^/]+/videos/(?:[^/]+/)?|video\.php\?v=)
            )
            (?P<id>[0-9]+)
            '''
    }

    _TESTS = [
        {
            'name': 'Get chat messages from past gaming broadcast',
            'params': {
                'url': 'https://www.facebook.com/disguisedtoast/videos/3629284013844544/',
                'max_messages': 100
            },
            'expected_result': {
                'messages_condition': lambda messages: 0 < len(messages) <= 100,
            }
        },

        {
            'name': 'Get chat messages from gaming clip',
            'params': {
                'url': 'https://www.facebook.com/disguisedtoast/videos/1170480696709027/',
                'max_messages': 100
            },
            'expected_result': {
                'messages_condition': lambda messages: 0 < len(messages) <= 100,
            }
        },

        {
            'name': 'Get chat messages from short gaming video',
            'params': {
                'url': 'https://www.facebook.com/disguisedtoast/videos/333201981735004/',
                'max_messages': 100
            },
            'expected_result': {
                'messages_condition': lambda messages: 0 < len(messages) <= 100,
            }
        },

        {
            'name': 'Get chat messages from long gaming video',
            'params': {
                'url': 'https://www.facebook.com/disguisedtoast/videos/918814568681983/',
                'start_time': 60,
                'end_time': 150
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },

        {
            'name': 'Get chat messages from video page',
            'params': {
                'url': 'https://www.facebook.com/video.php?v=570133851026337',
                'max_messages': 100
            },
            'expected_result': {
                'messages_condition': lambda messages: 0 < len(messages) <= 100,
            }
        },
        {
            'name': 'Get chat messages from short video',
            'params': {
                'url': 'https://www.facebook.com/338233632988842/videos/958020308373031',
                'max_messages': 100
            },
            'expected_result': {
                'messages_condition': lambda messages: 0 < len(messages) <= 100,
            }
        },
        {
            'name': 'Get chat messages from past broadcast',
            'params': {
                'url': 'https://www.facebook.com/kolaolootulive/videos/553617129001138/',
                'start_time': 567,
                'end_time': 1234
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },

        # Check for errors
        {
            'name': 'Video unavailable or private',
            'params': {
                'url': 'https://www.facebook.com/SRAVS.Gaming/videos/512714596679251/',
            },
            'expected_result': {
                'error': VideoUnavailable,
            }
        },
    ]
    _TESTS = [] # Remove support for facebook
    _GRAPH_API = _FB_HOMEPAGE + '/api/graphql/'

    def _graphql_request(self, program_params, retry_on_error=True, **post_kwargs):
        data = {
            'av': '0',
            '__user': '0',
            '__a': '1',
            '__comet_req': '1',
            'lsd': self.lsd,
            'server_timestamps': 'true',
            '__csr': '',
            'dpr': '1',
            '__ccg': 'MODERATE',
        }
        data.update(post_kwargs.pop('data', {}))
        post_kwargs['data'] = data

        max_attempts = program_params.get('max_attempts')
        for attempt_number in attempts(max_attempts):
            try:
                response = self._session_post(self._GRAPH_API, **post_kwargs)
                response_json = response.json()

                # Check for errors
                for error in response_json.get('errors') or []:
                    if error.get('code') == 1675004:
                        raise RateLimitError(f'Rate limit exceeded: {error}')

                return response_json

            except JSONDecodeError as e:
                self.retry(attempt_number, error=e, **program_params,
                           text=f'Unable to parse JSON: `{response.text}`')

            except RequestException as e:
                self.retry(attempt_number, error=e, **program_params)

            except RateLimitError as e:
                if retry_on_error:
                    self.retry(attempt_number, error=e, **program_params)
                else:
                    raise e

    _VIDEO_TITLE_REGEX = r'<meta\s+name=["\'].*title["\']\s+content=["\']([^"\']+)["\']\s*/>'

    def _get_initial_info(self, video_id, program_params):

        # Get metadata
        data = {
            'variables': json.dumps({
                'upNextVideoID': video_id,
            }),
            'doc_id': '4730353697015342'
        }

        json_data = self._graphql_request(program_params, data=data)

        video_data = multi_get(json_data, 'data', 'upNextVideoData')
        if not video_data:
            log('debug', json_data)
            raise VideoUnavailable('Video unavailable')

        return {
            'status': 'live' if video_data.get('is_live_streaming') else 'past',
            'video_type': 'video',
            'broadcast_status': video_data.get('broadcast_status'),
            'title': video_data.get('title_with_fallback'),
            'username': multi_get(video_data, 'owner', 'name'),
            'start_time': video_data.get('publish_time') * 1000000,
            'duration': video_data.get('playable_duration')
        }

    @staticmethod
    def _parse_feedback(feedback):
        new_feedback = {}

        edges = multi_get(feedback, 'top_reactions', 'edges')

        if not edges:
            return new_feedback

        new_feedback['reaction_types'] = []

        for edge in edges:
            node = edge.get('node')
            reaction_item = {
                'key': node.get('key'),
                'id': node.get('id'),
                'name': node.get('reaction_type'),
                'count': edge.get('reaction_count')
            }

            new_feedback['reaction_types'].append(reaction_item)

        new_feedback['total_count'] = multi_get(feedback, 'reactors', 'count')
        new_feedback['total_count_reduced'] = multi_get(
            feedback, 'reactors', 'count_reduced')

        return new_feedback

    @staticmethod
    def _get_text(item):
        return item.get('text') if item else None

    @staticmethod
    def _parse_image(item):
        return Image(item.get('uri'), item.get('width'), item.get('height')).json()

    @staticmethod
    def _get_uri(item):
        return item.get('uri')

    @staticmethod
    def _get_url(item):
        return item.get('url')

    @staticmethod
    def _parse_attachment_info(original_item):

        if isinstance(original_item, (list, tuple)) and len(original_item) > 0:
            original_item = original_item[0]

        if not original_item:
            return {}

        item = r.remap_dict(
            original_item, FacebookChatDownloader._TARGET_MEDIA_REMAPPING)

        # VideoTipJarPayment
        quantity = item.get('quantity')
        if quantity:
            item['text'] = f"Sent {quantity} Star{'s' if quantity != 1 else ''}"

        # For photos:
        blurred_image = item.pop('blurred_image', None)
        massive_image = item.pop('massive_image', None)

        if blurred_image and massive_image:
            item['text'] = Image(blurred_image, massive_image.get(
                'width'), massive_image.get('height')).json()

        # style_infos
        donation_comment_text = item.pop('donation_comment_text', None)
        if donation_comment_text:
            entity = multi_get(donation_comment_text,
                               'ranges', 0, 'entity') or {}

            item = r.remap_dict(
                entity, FacebookChatDownloader._TARGET_MEDIA_REMAPPING)
            item['text'] = donation_comment_text.get('text')

        # DEBUGGING
        original_type_name = original_item.get('__typename')
        if original_type_name not in FacebookChatDownloader._KNOWN_ATTACHMENT_TYPES:
            debug_log(
                f'Unknown attachment type: {original_type_name}',
                original_item,
                item
            )

        return item

    @staticmethod
    def _parse_target(media):
        item = {}

        return item

    @staticmethod
    def _parse_author_badges(item):

        keys = (('badge_asset', 'small'), ('information_asset', 'colour'))
        icons = list(map(lambda x: Image(
            FacebookChatDownloader._FB_HOMEPAGE + item.get(x[0]), 24, 24, x[1]).json(), keys))

        icons.append(
            Image(item.get('multiple_badge_asset'), 36, 36, 'large').json())

        return {
            'title': item.get('text'),
            'alternative_title': item.get('information_title'),
            'description': item.get('information_description'),
            'icons': icons,

            # badge_asset
            # multiple_badge_asset
            # information_asset

            'icon_name': item.get('identity_badge_type')

        }

    @staticmethod
    def _parse_attachment_renderer(item):
        attachment = multi_get(item, 'style_type_renderer',
                               'attachment') or item.get('attachment')
        if not attachment:
            debug_log(f'No attachment: {item}')
            return {}

        return FacebookChatDownloader._parse_attachment(attachment)

    @staticmethod
    def _parse_attachment(attachment):
        parsed = r.remap_dict(
            attachment, FacebookChatDownloader._ATTACHMENT_REMAPPING)

        for key in ('target', 'media', 'style_infos'):
            if parsed.get(key) == {}:
                parsed.pop(key)

        missing_keys = attachment.keys() - FacebookChatDownloader._KNOWN_ATTACHMENT_KEYS
        if missing_keys:
            debug_log(
                f'Missing attachment keys: {missing_keys}',
                attachment,
                parsed
            )

        return parsed

    _ATTACHMENT_REMAPPING = {
        'url': 'url',  # facebook redirect url,
        'source': r('source', _get_text),
        'title_with_entities': r('title', _get_text),

        'target': r('target', _parse_attachment_info),
        'media': r('media', _parse_attachment_info),
        'style_infos': r('style_infos', _parse_attachment_info),

        'attachment_text': r('text', _get_text),

        '__typename': 'type',

        'story_url': 'story_url',
        'story_attachment_link_renderer': r('story_attachment', _parse_attachment_renderer),

        'web_link': r('web_link', _get_url)
        # 'sticker_image': r('sticker_image', _get_uri),


    }

    _IGNORE_ATTACHMENT_KEYS = [
        'tracking',
        'action_links',
        'third_party_media_info'
    ]

    _KNOWN_ATTACHMENT_KEYS = set(
        list(_ATTACHMENT_REMAPPING.keys()) + _IGNORE_ATTACHMENT_KEYS)

    _TARGET_MEDIA_REMAPPING = {
        'id': 'id',
        '__typename': r('type', camel_case_split),
        'fallback_image': r('image', _parse_image),
        'is_playable': 'is_playable',
        'url': 'url',

        'mobileUrl': 'mobile_url',


        # Sticker
        'pack': 'pack',
        'label': 'label',
        'image': r('image', _parse_image),

        # VideoTipJarPayment

        'stars_image_on_star_quantity': 'icon',
        'spark_quantity': 'quantity',



        # Page
        'name': 'name',
        'category_name': 'category',
        'address': 'address',
        'overall_star_rating': 'overall_star_rating',

        'profile_picture': r('profile_picture', _get_uri),

        # Photo
        'accessibility_caption': 'accessibility_caption',

        'blurred_image': r('blurred_image', _get_uri),
        'massive_image': 'massive_image',


        # FundraiserForStoryDonationAttachmentStyleInfo
        'donation_comment_text': 'donation_comment_text'

    }

    _KNOWN_ATTACHMENT_TYPES = [
        'Sticker',
        'VideoTipJarPayment',
        'Page',
        'Video',

        'Group',
        'ProfilePicAttachmentMedia',
        'User',
        'Photo',

        'ExternalUrl',
        'GenericAttachmentMedia',

        'ChatCommandResult',

        'CommentMessageInfo',
        'FundraiserForStoryDonationAttachmentStyleInfo',

        'Event'
    ]

    _REMAPPING = {
        'id': 'message_id',
        'community_moderation_state': 'community_moderation_state',

        # attachments

        'author': 'author',


        'feedback': r('reactions', _parse_feedback),
        'created_time': r('timestamp', lambda x: x * 1000000),


        'upvote_downvote_total': 'upvote_downvote_total',
        'is_author_banned_by_content_owner': 'is_author_banned',
        'is_author_original_poster': 'is_author_original_poster',
        'is_author_bot': 'is_author_bot',
        'is_author_non_coworker': 'is_author_non_coworker',
        # if banned, ban_action?

        'comment_parent': 'comment_parent',

        'edit_history': r('number_of_edits', lambda x: x.get('count')),


        'timestamp_in_video': 'timestamp_in_video',
        'written_while_video_was_live': 'written_while_video_was_live',



        'translatability_for_viewer': r('message_dialect', lambda x: x.get('source_dialect_name')),


        'url': 'message_url',

        'body': r('message', _get_text),

        'identity_badges_web': r('author_badges', lambda x: list(map(FacebookChatDownloader._parse_author_badges, x))),

        'attachments': r('attachments', lambda x: list(map(FacebookChatDownloader._parse_attachment_renderer, x)))

    }

    _AUTHOR_REMAPPING = {
        'id': 'id',
        'name': 'name',
        '__typename': r('type', camel_case_split),
        'url': 'url',

        'is_verified': 'is_verified',

        'gender': r('gender', lambda x: x.lower()),
        'short_name': 'short_name'
    }

    @ staticmethod
    def _parse_node(node, parse_time=False, start_time=None):
        info = r.remap_dict(node, FacebookChatDownloader._REMAPPING)

        author_info = info.pop('author', {})
        BaseChatDownloader._move_to_dict(
            info, 'author', create_when_empty=True)

        info['author'] = r.remap_dict(
            author_info, FacebookChatDownloader._AUTHOR_REMAPPING)

        if 'profile_picture_depth_0' in author_info:
            info['author']['images'] = []
            for size in ((0, 32), (1, 24)):
                url = multi_get(
                    author_info, f'profile_picture_depth_{size[0]}', 'uri')
                info['author']['images'].append(
                    Image(url, size[1], size[1]).json())

        # author_badges = info.pop('author_badges', None)
        # if author_badges:
        #     info['author']['badges'] = author_badges

        in_reply_to = info.pop('comment_parent', None)
        if isinstance(in_reply_to, dict) and in_reply_to:
            info['in_reply_to'] = FacebookChatDownloader._parse_node(
                in_reply_to, parse_time, start_time)

        if parse_time:
            timestamp_in_video = info.get('timestamp_in_video')
            if timestamp_in_video is not None:
                if start_time is None:
                    info['time_in_seconds'] = timestamp_in_video
                else:
                    info['time_in_seconds'] = (
                        info['timestamp'] - start_time)/1e6

                info['time_text'] = seconds_to_time(info['time_in_seconds'])

        message = info.get('message')
        if message:
            info['message'] = message
            info['message_type'] = 'text_message'
        else:
            info.pop('message', None)  # remove if empty

        # remove the following if empty:
        if info.get('reactions') == {}:
            info.pop('reactions')
        if info.get('attachments') == []:
            info.pop('attachments')

        return info

    def _get_live_chat_messages_by_video_id(self, video_id, params):
        buffer_size = 25  # max num comments returned by api call
        # cursor = ''
        variables = {
            'videoID': video_id
        }
        data = {
            'variables': json.dumps(variables),
            'doc_id': '4889623951078943',  # specifies what API call this is?
            # 'cursor' : cursor
            # &first=12&after=<end_cursor>
        }

        first_try = True

        last_ids = []
        while True:

            json_data = self._graphql_request(params, data=data)

            feedback = multi_get(json_data, 'data', 'video', 'feedback')
            if not feedback:
                log('debug', f'No feedback: {json_data}')
                continue

            top_level_comments = multi_get(
                json_data, 'data', 'video', 'feedback', 'top_level_comments')

            errors = json_data.get('errors')
            if errors:
                # TODO will usually resume getting chat..
                # maybe add timeout?
                log('debug', f'Errors detected: {errors}')
                continue

            if not top_level_comments:
                log('debug', f'No top level comments: {json_data}')
                continue

            # Parse items:
            parsed_items = []
            for edge in top_level_comments.get('edges') or []:
                node = edge.get('node')
                if not node:
                    log('debug', f'No node found in edge: {edge}')
                    continue
                parsed_items.append(FacebookChatDownloader._parse_node(node))

            # Sort items
            parsed_items.sort(key=lambda x: x['timestamp'])

            # TODO - get pagination working
            # page_info = top_level_comments.get('page_info')
            # after = page_info.get('end_cursor')
            # has_next_page = page_info.get('has_next_page')

            num_to_add = 0
            for item in parsed_items:
                comment_id = item.get('message_id')

                # remove items that have already been parsed
                if comment_id in last_ids:
                    continue

                last_ids.append(comment_id)
                last_ids = last_ids[-buffer_size:]  # force x items

                # TODO determine whether to add or not (message types/groups)

                num_to_add += 1
                yield item

            if num_to_add == 0:
                time_to_sleep = 1
                log('debug',
                    f'No actions, sleeping for {time_to_sleep} seconds')
                interruptible_sleep(time_to_sleep)

            # got 25 items, and this isn't the first one
            if num_to_add >= buffer_size and not first_try:
                log(
                    'warning',
                    'Messages may be coming in faster than requests are being made.'
                )

            if first_try:
                first_try = False


    def _get_chat_from_vod(self, feedback_id, stream_start_time, end_time, params):
        # method 1 - only works for vods. Guaranteed to get all, but can't choose start time
        # ordered by timestamp

        data = {
            'fb_api_req_friendly_name': 'CometUFICommentsProviderPaginationQuery',
            'doc_id': '4310877875602018'
        }

        ordering = 'LIVE_STREAMING'  # 'TOPLEVEL' 'RANKED_THREADED' 'CHRONOLOGICAL'

        variables = {
            'feedLocation': 'TAHOE',
            'feedbackID': feedback_id,
            'feedbackSource': 41,
            'last': 50,  # max step is 50
            'includeHighlightedComments': True,
            'includeNestedComments': True,
            'initialViewOption': ordering,
            'topLevelViewOption': ordering,
            'viewOption': ordering
        }

        before = None
        while True:
            variables['before'] = before
            variables['isPaginating'] = before is not None

            data['variables'] = json.dumps(variables)

            json_data = self._graphql_request(params, retry_on_error=False, data=data)

            info = multi_get(json_data, 'data', 'feedback')
            if not info:
                log('debug', f'No feedback: {json_data}')
                break

            display_comments = info.get('display_comments')
            edges = display_comments.get('edges') or []

            parsed_items = []
            for edge in reversed(edges):
                node = edge.get('node')
                if not node:
                    log('debug', f'No node found in edge: {edge}')
                    continue

                parsed = FacebookChatDownloader._parse_node(
                    node, True, stream_start_time)

                time_in_seconds = parsed.get('time_in_seconds')
                if time_in_seconds is None:
                    continue
                if time_in_seconds > end_time:
                    return
                parsed_items.append(parsed)

            parsed_items.sort(key=lambda x: x['timestamp'])
            yield from parsed_items

            page_info = display_comments.get('page_info') or {}
            if not page_info.get('has_previous_page'):
                break

            before = page_info.get('start_cursor')

    def _get_chat_from_video(self, feedback_id, start_time, end_time, params):
        # method 2 - works for all videos, but sometimes misses comments
        # max messages is 30 per minute
        # ordered by time_in_seconds
        log('debug', 'Running method 2')

        data = {
            'fb_api_req_friendly_name': 'CometLiveVODCommentListRefetchableQuery',
            'doc_id': '3941623715965411'
        }

        # By default, Facebook gets messages by the minute
        time_increment = 600  # 10 mins

        next_start_time = max(start_time, 0)

        variables = {
            'id': feedback_id
        }

        while True:
            next_end_time = min(next_start_time + time_increment, end_time)

            variables['afterTime'] = next_start_time
            variables['beforeTime'] = next_end_time

            data['variables'] = json.dumps(variables)

            json_data = self._graphql_request(params, data=data)

            edges = multi_get(json_data, 'data', 'node',
                              'video_timestamped_comments', 'edges') or []

            for edge in edges:
                node = edge.get('node')
                if not node:
                    log('debug', f'No node found in edge: {edge}')
                    continue
                yield FacebookChatDownloader._parse_node(node, True)

            if next_end_time >= end_time:
                return
            next_start_time = next_end_time

    def _get_chat_replay_messages_by_video_id(self, video_id, max_duration, initial_info, params):
        feedback_id = base64_encode(f'feedback:{video_id}')
        broadcast_status = initial_info.get('broadcast_status')
        stream_start_time = initial_info.get('start_time')

        start_time = ensure_seconds(params.get('start_time'), 0)
        end_time = min(ensure_seconds(params.get(
            'end_time'), float('inf')), max_duration)

        success = False

        if broadcast_status == 'VOD_READY' and params.get('start_time') is None:
            # no start time, get until end_time
            log('debug', 'Running method 1')
            try:
                for i in self._get_chat_from_vod(feedback_id, stream_start_time, end_time, params):
                    time_in_seconds = i.get('time_in_seconds')
                    if time_in_seconds is not None:
                        start_time = max(start_time, time_in_seconds)
                    yield i
                success = True

            except RateLimitError as e:
                pass

        if not success:  # Fallback
            yield from self._get_chat_from_video(feedback_id, start_time, end_time, params)

    def _get_chat_by_video_id(self, match, params):
        return self.get_chat_by_video_id(match.group('id'), params)

    def get_chat_by_video_id(self, video_id, params):

        initial_info = self._get_initial_info(video_id, params)

        start_time = params.get('start_time')
        end_time = params.get('end_time')

        # if start or end time specified, use chat replay...
        # The tool works for both active and finished live streams.
        # if start/end time are specified, vods will be prioritised
        # if is live stream and no start/end time specified
        if initial_info.get('status') == 'live' and not start_time and not end_time:
            generator = self._get_live_chat_messages_by_video_id(
                video_id, params)
        else:
            max_duration = initial_info.get('duration', float('inf'))

            generator = self._get_chat_replay_messages_by_video_id(
                video_id, max_duration, initial_info, params)

        return Chat(
            generator,
            id=video_id,
            **initial_info
        )

    _STREAM_PAGE = 'https://www.facebook.com/gaming/browse/live/?s=VIEWERS&language=ALL_LANG'

    def generate_urls(self, **kwargs):
        yield from self._generate_live(kwargs.get('livestream_limit'), **kwargs)
        yield from self._generate_videos(kwargs.get('vod_limit'), **kwargs)
        yield from self._generate_clips(kwargs.get('clip_limit'), **kwargs)

    def _generate_live(self, limit, **kwargs):
        # https://www.facebook.com/gaming/browse/live/?s=VIEWERS&language=ALL_LANG
        return self._generate_urls('live', limit, **kwargs)

    def _generate_videos(self, limit, **kwargs):
        # https://www.facebook.com/gaming/videos
        return self._generate_urls('videos', limit, **kwargs)

    def _generate_clips(self, limit, **kwargs):
        # https://www.facebook.com/gaming/clips/
        return self._generate_urls('clips', limit, **kwargs)

    def _generate_urls(self, video_type, limit, **kwargs):
        max_attempts = 10
        program_params = {
            'max_attempts': max_attempts
        }

        step = 8
        if video_type in ('live', 'videos'):
            variables = {
                'count': step,
                'params': {
                    'following': None,  # False
                    'game_id': None,
                    'language': 'ALL_LANG',
                    'remove_following': True,
                    'sort_order': 'VIEWERS'  # None 'SUGGESTED'
                }
            }

            key = 'top_live' if video_type == 'live' else 'top_was_live'

        else:  # video_type == 'clips':
            variables = {
                'count': step,
                'following': False,
                'game_id': None,
                'streamer_id': None,
                'sort_order': 'VIEWERS'  # None 'SUGGESTED'
            }
            key = 'top_weekly_clips'

        doc_ids = {
            'live': 3843810065738698,
            'videos': 4591277870888795,
            'clips': 3586924904747093
        }
        doc_id = doc_ids.get(video_type)

        data = {
            'doc_id': doc_id
        }

        count = 0
        while True:
            data['variables'] = json.dumps(variables)

            json_data = self._graphql_request(program_params, data=data)

            top_live = multi_get(json_data, 'data', 'gaming_video', key)

            if not top_live:
                log('debug', f'No data found: {json_data}')
                return

            edges = top_live.get('edges') or []

            for edge in edges:
                if count >= limit:
                    return

                url = multi_get(edge, 'node', 'url')
                if url:
                    yield url
                    count += 1

            page_info = top_live.get('page_info')

            has_next_page = page_info.get('has_next_page')

            if not has_next_page:
                break
            variables['cursor'] = page_info.get('end_cursor')




import requests
from http.cookiejar import (MozillaCookieJar, Cookie)
import os
import re
from json import JSONDecodeError

from ..errors import (
    InvalidParameter,
    RetriesExceeded,
    CookieError,
    UnexpectedError,
    URLNotProvided,
    SiteNotSupported,
    InvalidURL
)

from ..utils.core import (
    get_title_of_webpage,
    pause,
    safe_print,
    safe_path
)

from ..utils.timed_utils import (
    timed_input,
    interruptible_sleep
)
from ..debugging import log


class Image():
    def __init__(self, url, width=None, height=None, image_id=None):
        """Create an Image object

        :param url: The URL of the actual image
        :type url: str
        :param width: The width of the image, defaults to None
        :type width: int, optional
        :param height: The height of the image, defaults to None
        :type height: int, optional
        :param image_id: A identifier for the image, usually of the form: {width}x{height}, defaults to None
        :type image_id: str, optional
        """
        self.url = url

        if self.url.startswith('//'):
            self.url = 'https:' + self.url

        self.width = width
        self.height = height

        if width and height and not image_id:
            self.id = f'{width}x{height}'
        elif image_id:
            self.id = image_id

    def json(self):
        """Return the JSON representation of an Image

        :return: JSON representation of the object
        :rtype: dict
        """
        return {k: v for k, v in self.__dict__.items() if v is not None}


class Remapper():
    """Class used to control the remapping of one dictionary to another dictionary."""

    def __init__(self, new_key=None, remap_function=None, to_unpack=False):
        """Create a Remapper object

        :param new_key: The new key of the item, defaults to None
        :type new_key: str, optional
        :param remap_function: The remapping function, defaults to None
        :type remap_function: function, optional
        :param to_unpack: Unpack the remapped item (to map to multiple output keys),
            defaults to False
        :type to_unpack: bool, optional
        :raises ValueError: if unable to perform a remapping
        """

        if new_key is not None and to_unpack:
            # New key is specified, but must unpack. Not allowed
            raise ValueError(
                'If to_unpack is True, new_key may not be specified.')

        self.new_key = new_key

        if isinstance(remap_function, staticmethod):
            remap_function = remap_function.__func__

        if remap_function is None or not (hasattr(remap_function, '__call__')):
            raise ValueError('remap_function must be callable or None.')

        self.remap_function = remap_function
        self.to_unpack = to_unpack

    @staticmethod
    def remap(info, remapping_dict, remap_key, remap_input, keep_unknown_keys=False, replace_char_with_underscores=None):
        """A function used to remap items from one dictionary to another

        :param info: Output dictionary
        :type info: dict
        :param remapping_dict: Dictionary of remappings
        :type remapping_dict: dict
        :param remap_key: The key of the remapping
        :type remap_key: str
        :param remap_input: The input sent to the remapping function
        :type remap_input: object
        :param keep_unknown_keys: If no remapping is found, keep the data
            with its original key and value. Defaults to False
        :type keep_unknown_keys: bool, optional
        :param replace_char_with_underscores: If no remapping is found,
            replace a character in the key with underscores. Defaults to None
        :type replace_char_with_underscores: str, optional
        :raises ValueError: if attempting to unpack an item that is not a dictionary,
            or if an unknown remapping is specified
        """

        remap = remapping_dict.get(remap_key)

        if remap:  # A matching 'remapping' has been found, apply this remapping
            if isinstance(remap, Remapper):
                new_key = remap.new_key  # or remap_key

                # Perform transformation
                if remap.remap_function:  # Has a remap function
                    new_value = remap.remap_function(remap_input)
                else:  # No remap function specified, apply identity transformation
                    new_value = remap_input

                # Assign values to info
                if not remap.to_unpack:
                    info[new_key] = new_value
                elif isinstance(new_value, dict):
                    info.update(new_value)
                else:
                    raise ValueError(
                        'Unable to unpack item which is not a dictionary.')

            elif isinstance(remap, str):
                # If it is just a string, simply assign the new value to this key
                info[remap] = remap_input
            else:
                raise ValueError('Unknown remapping specified.')

        elif keep_unknown_keys:
            if replace_char_with_underscores:
                remap_key = remap_key.replace(
                    replace_char_with_underscores, '_')
            info[remap_key] = remap_input

    @staticmethod
    def remap_dict(input_dictionary, remapping_dict, keep_unknown_keys=False, replace_char_with_underscores=None):
        """Given an input dictionary and a remapping dictionary, return the remapped dictionary

        :param input_dictionary: Input dictionary
        :type input_dictionary: dict
        :param remapping_dict: Dictionary of Remapper objects
        :type remapping_dict: dict
        :param keep_unknown_keys: If no remapping is found, keep the data
            with its original key and value. Defaults to False
        :type keep_unknown_keys: bool, optional
        :param replace_char_with_underscores: If no remapping is found,
            replace a character in the key with underscores. Defaults to None
        :type replace_char_with_underscores: str, optional
        :return: Remapped dictionary
        :rtype: dict
        """

        info = {}
        for key in input_dictionary:
            Remapper.remap(
                info, remapping_dict, key, input_dictionary[key],
                keep_unknown_keys=keep_unknown_keys,
                replace_char_with_underscores=replace_char_with_underscores
            )
        return info


class SiteDefault:
    """Allows for sites to specify default parameters. Additionally, different
    sites can specify different values for the same input parameter."""

    def __init__(self, name):
        """Create a SiteDefault object

        :param name: The key which will be checked in the `_SITE_DEFAULT_PARAMS`
        dictionary to get the site's default value
        :type name: str
        """
        self.name = name


class Chat():
    """Class used to manage all chat data for a single stream or video.

    Classes that extend `BaseChatDownloader` contain the `get_chat` method,
    which returns a `Chat` object. These objects are iterable, where the
    next value is yielded from the object's `chat` generator method.
    """

    def __init__(self, chat=None, title=None, duration=None, status=None, video_type=None, start_time=None, id=None, **kwargs):
        """Create a Chat object

        :param chat: Generator method for retrieving chat messages, defaults to None
        :type chat: generator, optional
        :param title: Stream or video title, defaults to None
        :type title: str, optional
        :param duration: Duration of the stream or video, defaults to None
        :type duration: float, optional
        :param status: Status of the stream or video (e.g., live, upcoming, past),
            defaults to None
        :type status: str, optional
        :param video_type: Type of the stream or video (e.g., video, premiere, clip),
            defaults to None
        :type video_type: str, optional
        :param start_time: Start time of the stream (or upload date of video)
            in UNIX microseconds, defaults to None
        :type start_time: float, optional
        """

        self.chat = chat

        self.title = title
        self.duration = duration

        self.status = status
        self.video_type = video_type

        self.start_time = start_time
        self.id = id

        # TODO
        # author/user/uploader/creator

        self._output_writer = None
        self._output_callback = None

    def __iter__(self):
        """Allows the object to be iterable

        :return: This object
        :rtype: Chat
        """
        return self

    def _init_writer(self):
        if self._output_writer.is_initialised():
            return  # Ignore if writer is already initialised

        # Special formatting of output name:
        # Allowed keys are specified here
        # Remove invalid characters from output file name
        self._output_writer.file_name = self._output_writer.file_name.format(
            title=safe_path(self.title),
            id=safe_path(self.id)
        )

        log('debug', f'Writing to file: {self._output_writer.file_name}')
        # Only actually initialise here
        self._output_writer._real_init()

        if self._output_writer.is_default():
            self._output_callback = lambda item: self._output_writer.write(
                self.format(item), flush=True)
        else:
            self._output_callback = lambda item: self._output_writer.write(
                item, flush=True)

    def attach_writer(self, writer):
        # writer is a ContinuousWriter
        self._output_writer = writer

    def __next__(self):
        """Get the next chat message from the generator

        :return: The next chat item
        :rtype: dict
        """
        try:
            item = next(self.chat)

            if self._output_writer is not None:  # writer has been attached
                self._init_writer()

            if self._output_callback is not None:  # output callback
                self._output_callback(item)

            return item
        except StopIteration as e:
            # Safely close output file when done
            if self._output_writer is not None:
                self._output_writer.close()
            raise e

    def print_formatted(self, item, flush=True):
        """Safely print the formatted message

        :param item: The chat item to be printed
        :type item: dict
        """
        safe_print(self.format(item), flush=flush)

    def format(self, item):
        """Format chat messages

        :param item: The chat item to be formatted
        :type item: dict
        :raises NotImplementedError: if this method has not been overridden later
        """
        raise NotImplementedError


class BaseChatDownloader:
    """Base class for chat downloaders. Each supported site should have its
    own chat downloader. Subclasses should redefine the `_VALID_URLS`
    dictionary which creates a mapping between functions and their matching
    regular expressions. Optionally, subclasses should also redefine `_NAME`,
    `_SITE_DEFAULT_PARAMS` and `_TESTS` fields."""

    _NAME = None

    _SITE_DEFAULT_PARAMS = {
        # MAY NOT specify message_types. must always be empty
        'message_groups': ['messages'],
        'format': 'default',
    }

    # For general tests (non-site specific)
    _TESTS = [
        {
            'name': 'Inactivity timeout',
            'params': {
                'url': 'https://twitch.tv/xenova',
                'inactivity_timeout': 5,
                'timeout': 20,  # As a fallback
            }
        },
        {
            'name': 'Get a certain number of messages from a livestream.',
            'params': {
                'url': 'https://www.youtube.com/watch?v=5qap5aO4i9A',
                'max_messages': 10,
                'timeout': 60,  # As a fallback
            },

            'expected_result': {
                'messages_condition': lambda messages: len(messages) <= 10,
            }
        },


        {
            'name': 'Scheme not supplied',
            'params': {
                'url': 'www.youtube.com/watch?v=5qap5aO4i9A',
                'max_messages': 10,
                'timeout': 60,  # As a fallback
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) <= 10,
            }
        },

        # Tests for errors
        {
            'name': 'No URL provided.',
            'params': {
                'url': '',
            },
            'expected_result': {
                'error': URLNotProvided
            }
        },
        {
            'name': 'Site not supported',
            'params': {
                'url': 'https://www.example.com',
            },
            'expected_result': {
                'error': SiteNotSupported
            }
        },
        {
            'name': 'Invalid URL',
            'params': {
                'url': '#',
            },
            'expected_result': {
                'error': InvalidURL
            }
        }
    ]

    @staticmethod
    def _must_add_item(item, message_groups_dict, messages_groups_to_add, messages_types_to_add):

        # Force mutual exclusion
        if messages_types_to_add:
            # messages_types is set
            messages_groups_to_add = []

        if 'all' in messages_groups_to_add or 'all' in messages_types_to_add:  # user wants everything
            return True

        valid_message_types = []
        for message_group in messages_groups_to_add or []:
            valid_message_types += message_groups_dict.get(message_group, [])

        for message_type in messages_types_to_add or []:
            valid_message_types.append(message_type)

        return item.get('message_type') in valid_message_types

    def __init__(self,
                 **kwargs
                 ):
        """Initialise a session with various parameters

        :raises CookieError: if unable to read or parse the cookie file
        """

        # Start a new session
        self.session = requests.Session()

        headers = kwargs.get('headers')
        if headers is None:
            headers = {
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36',
                'Accept-Language': 'en-US, en, *'  # 'de-CH'#'fr-CH'#
            }
        self.session.headers = headers

        # Set proxies if present
        proxy = kwargs.get('proxy')
        if proxy is not None:
            if proxy == '':
                proxies = {}
            else:
                proxies = {'http': proxy, 'https': proxy}

            self.session.proxies.update(proxies)

        # Set cookies if present
        cookies = kwargs.get('cookies')
        cj = MozillaCookieJar(cookies)

        if cookies:  # is not None
            # Only attempt to load if the cookie file exists.
            if os.path.exists(cookies):
                cj.load(ignore_discard=True, ignore_expires=True)
            else:
                raise CookieError(
                    f'The file "{cookies}" could not be found.')
        self.session.cookies = cj

    def get_session_headers(self, key):
        return self.session.headers.get(key)

    def update_session_headers(self, new_headers):
        self.session.headers.update(new_headers)

    def clear_cookies(self):
        """Clear the session's cookies."""
        self.session.cookies.clear()

    def _get_cookies_dict(self):
        """Returns a key/value dictionary from the session's CookieJar

        :return: Dictionary of cookies
        :rtype: dict
        """
        return requests.utils.dict_from_cookiejar(self.session.cookies)

    def set_cookie_value(self, domain, name, value, expire_time=None, port=None,
                         path='/', secure=False, discard=False, rest={}, **kwargs):
        cookie = Cookie(
            0, name, value, port, port is not None, domain, True,
            domain.startswith('.'), path, True, secure, expire_time,
            discard, None, None, rest)
        self.session.cookies.set_cookie(cookie)

    def get_cookie_value(self, name, default=None):
        """Return the value for key if key is in the cookie dictionary, else default.

        :param name: The key of the cookie
        :type name: str
        :param default: Return this value if the specified cookie cannot be found, defaults to None
        :type default: object, optional
        :return: The cookie value, or default
        :rtype: Union[str, object, None]
        """
        return self._get_cookies_dict().get(name, default)

    def close(self):
        """Close the session. Once this has been called, no more requests can be made."""
        self.session.close()
        log('debug', 'Session closed.')

    def _session_post(self, url, **kwargs):
        """Make a post request using the current session."""
        return self.session.post(url, **kwargs)

    def _session_get(self, url, **kwargs):
        """Make a get request using the current session."""
        return self.session.get(url, **kwargs)

    def _session_get_json(self, url, **kwargs):
        """Make a get request using the current session and return as JSON."""
        return self._session_get(url, **kwargs).json()

    def get_site_value(self, value):
        """Get the site's default value for a certain parameter

        :param value: The value
        :type value: Union[SiteDefault, object]
        :return: The site's default value
        :rtype: object
        """
        if isinstance(value, SiteDefault):
            return self._SITE_DEFAULT_PARAMS.get(
                value.name, BaseChatDownloader._SITE_DEFAULT_PARAMS.get(value.name))
        else:
            return value

    _VALID_URLS = {
        # function_name: regex
    }

    @classmethod
    def matches(cls, url):
        """Used to check if a url matches any of the
        regular expressions specified in the classes
        `_VALID_URLS` dictionary.

        :return: If a match is found, the function name and
            match object is returned, otherwise None.
        :rtype: (str, re.Match)
        """
        for function_name, regex in cls._VALID_URLS.items():

            if isinstance(regex, str):
                match = re.search(regex, url)
                if match:
                    return function_name, match

        return None

    def generate_urls(self, **kwargs):
        """This method should be implemented in a subclass and should return
        a generator which yields URLs for testing.

        :raises NotImplementedError: if not implemented and called from a subclass
        """
        raise NotImplementedError

    @staticmethod
    def _move_to_dict(info, dict_name, replace_key=None, create_when_empty=False, *info_keys):
        """
        Move all items with keys that contain some text to a separate dictionary.

        These keys are modifed by removing some text.
        """
        if replace_key is None:
            replace_key = dict_name + '_'

        new_dict = {}

        for key in (info_keys or info or {}).copy():
            if replace_key in key:
                info_item = info.pop(key, None)
                new_key = key.replace(replace_key, '')

                # set it if it contains info
                if info_item not in (None, [], {}):
                    new_dict[new_key] = info_item

        if dict_name in info:
            info[dict_name].update(new_dict)
        elif create_when_empty or new_dict != {}:  # dict_name not in info
            info[dict_name] = new_dict

        return new_dict

    @staticmethod
    def retry(attempt_number, max_attempts=1, error=None, retry_timeout=None, text=None, interruptible_retry=True, **kwargs):
        """Retry to occur after an error occurs

        :param attempt_number: The current attempt number
        :type attempt_number: int
        :param max_attempts: The maximum number of attempts allowed
        :type max_attempts: int, optional
        :param error: The error which was raised, defaults to None
        :type error: Exception, optional
        :param retry_timeout: The number of seconds to sleep after failing,
            defaults to None (i.e. use exponential back-off)
        :type retry_timeout: float, optional
        :param text: Items to display on retry, defaults to None
        :type text: object, optional
        :raises RetriesExceeded: if the maximum number of retries has been exceeded
        """
        if attempt_number >= max_attempts:
            raise RetriesExceeded(
                f'Maximum number of retries has been reached ({max_attempts}).')

        if text is None:
            text = []
        elif not isinstance(text, (tuple, list)):
            text = [text]

        if retry_timeout is None:  # use exponential backoff
            if attempt_number > 1:
                time_to_sleep = 2**(attempt_number - 2)
            else:
                time_to_sleep = 0

        elif isinstance(retry_timeout, (int, float)):  # valid timeout value
            time_to_sleep = retry_timeout
        else:
            time_to_sleep = -1  # wait for user input

        must_sleep = time_to_sleep >= 0
        if must_sleep:
            if interruptible_retry:
                sleep_text = f'(sleep for {time_to_sleep}s or press Enter)'
            else:
                sleep_text = f'(sleep for {time_to_sleep}s)'
        else:
            sleep_text = ''

        retry_text = f'Retry #{attempt_number} {sleep_text}.'

        if isinstance(error, Exception):
            retry_text += f' {error} ({error.__class__.__name__})'

        if isinstance(error, JSONDecodeError):
            log('debug', error.__dict__)
            page_title = get_title_of_webpage(error.doc)
            if page_title:
                log('debug', f'Title: {page_title}')

        log('warning', text + [retry_text])

        if must_sleep:
            if interruptible_retry:
                timed_input(time_to_sleep)
            else:
                interruptible_sleep(time_to_sleep)
        else:
            pause()

    @staticmethod
    def check_for_invalid_types(messages_types_to_add, allowed_message_types):
        """Used to check for invalid message types

        :param messages_types_to_add: List of message types to add
        :type messages_types_to_add: list
        :param allowed_message_types: List of allowed message type
        :type allowed_message_types: list
        :raises InvalidParameter: if invalid types are specified
        """
        invalid_types = set(messages_types_to_add) - set(allowed_message_types)
        if invalid_types:
            raise InvalidParameter(f'Invalid types specified: {invalid_types}')

    @staticmethod
    def get_mapped_keys(remapping):
        mapped_keys = set()
        for key in remapping:
            value = remapping[key]

            if isinstance(value, Remapper):
                value = value.new_key
            mapped_keys.add(value)

        return mapped_keys
